"""

    Useful functions and classes, mostly relating to the Pattern class
    but used in multiple sub-packages of FoxDot.

"""

from __future__ import absolute_import, division, print_function

import sys
import json
import itertools
from socket import timeout as socket_timeout

try:
    from urllib.request import urlopen
    from urllib.error import URLError
except ImportError:
    from urllib2 import urlopen, URLError

# Functions

def get_pypi_version():
    """ Returns the most up-to-date version number on PyPI. Return None on error """
    try:
        addr = "https://pypi.org/pypi/FoxDot/json"
        page = urlopen(addr, timeout=2.5)
        data = json.loads(page.read().decode("utf-8"))
        version = data["info"]["version"]
    except (URLError, socket_timeout):
        version = None
    return version


def stdout(*args):
    """ Forces prints to stdout and not console """
    sys.__stdout__.write(" ".join([str(s) for s in args]) + "\n")

def sliceToRange(s):
    start = s.start if s.start is not None else 0
    stop  = s.stop
    step  = s.step if s.step is not None else 1
    try:
        return list(range(start, stop, step))
    except OverflowError:
        raise TypeError("range() integer end argument expected, got NoneType")

def LCM(*args):
    """ Lowest Common Multiple """

    args = [n for n in args if n != 0]

    # Base case
    if len(args) == 0:
        return 1

    elif len(args) == 1:
        return args[0]

    X = list(args)

    while any([X[0]!=K for K in X]):

        i = X.index(min(X))
        X[i] += args[i]

    return X[0]

def EuclidsAlgorithm(n, k, lo=0, hi=1):

    if n == 0: return [n for i in range(k)]

    data = [[hi if i < n else lo] for i in range(k)]

    while True:

        k = k - n

        if k <= 1:
            break

        elif k < n:
            n, k = k, n

        for i in range(n):
            data[i] += data[-1]
            del data[-1]

    return [x for y in data for x in y]


def PulsesToDurations(data):
    """ Returns a list of durations based on pulses (1s) and blanks (0s).
        Data should be a list of [1,0] where 1 is a pulse. """
    
    count, seq = 1, []

    for item in data[1:]:

        if item == 1:

            seq.append(count)
            count = 1

        else:

            count += 1

    seq.append(count)

    return seq

def get_first_item(array):
    """ Returns first item from a possibly nested list"""
    try:
        return get_first_item(array[0])
    except (TypeError, IndexError):
        return array

def modi(array, i, debug=0):
    """ Returns the modulo index i.e. modi([0,1,2],4) will return 1 """
    try:
        return array[i % len(array)]
    except(TypeError, AttributeError, ZeroDivisionError):
        return array

def get_expanded_len(data):
    """ (0,(0,2)) returns 4. int returns 1 """
    if type(data) is str and len(data) == 1:
        return 1
    l = []
    try:
        for item in data:
            try:
                l.append(get_expanded_len(item))
            except(TypeError, AttributeError):
                l.append(1)
        return LCM(*l) * len(data)
    except TypeError:
        return 1

def max_length(*patterns):
    """ Returns the largest length pattern """
    return max([len(p) for p in patterns])

def get_inverse_op(method):
    """ Returns the opposite __dunder__ method e.g.
        get_inverse_op("__add__") -> "__radd__"
        get_inverse_op("__ror__") -> "__or__"
    """
    if method.startswith("__r"):
        return method.replace("__r", "__")
    elif method.startswith("__"):
        return method.replace("__", "__r", 1)
    return method

def isiterable(obj):
    """ Returns true if an object is iterable by using `iter(obj)`"""
    try:
        iter(obj)
        return True
    except:
        return False

def recursive_any(seq):
    """ Like any but checks lists recursively """
    for item in seq:
        if isiterable(item):
            if recursive_any(item):
                return True
        else:
            if bool(item):
                return True
    else:
        return False

# Classes

class dots:
    """ Class for representing long Patterns in strings """
    def __repr__(self):
        return '...'




#!/usr/bin/python
""" Collection of functions used in the Workspace package """

from __future__ import absolute_import, division, print_function
from .Format import *
import sys

def stdout(*args):
    """ Forces prints to stdout and not console """
    sys.__stdout__.write(" ".join([str(s) for s in args]) + "\n")

def index(index1, index2=None):
    if type(index1) == str and index2 == None:
        return tuple(int(n) for n in index1.split('.'))
    return str(index1) + '.' + str(index2)

def empty(string):
    return len(string.replace(' ','')) == 0

def isDelete(char):
    return char == "\x08"

def isReturn(char):
    return char == "\r"

def isHex(char):
    return len(repr(char)) > 3

def indented(s):
    if len(s) == 0:
        return 0
    for i, char in enumerate(s):
        if char != " ":
            break
    return i

def at_function(index, string):
    """ Returns true if the index is at a colon in the string """
    return False if index == 0 else string[index-1] == ":"

def in_brackets(index, string):
    """ Returns true if index is between a pair of brackets (could be non closing) """
    
    left_b = list("({[")
    right_b = list(")}]")

    count = dict([(l,0) for l in left_b])

    index = index - 1 # Look to the character to the left to start

    while index > 0:

        char = string[index]

        if char in left_b:

            if count[char] == 0:

                return True

            else:

                count[char] -= 1

        elif char in right_b:

            count[brackets[char]] += 1

        index -= 1

    return False

def get_tabspace(s):
    """ Returns the amount of whitespace at the start of a string """
    return indented(s) * " "

def open_bracket(text):
    """ Returns the index of the last open bracket """

    count = dict([(l,0) for l in left_b])

    for i, char in enumerate(text):

        if   char in left_b:

            count[char] += 1

        elif char in right_b:

            count[brackets[char]] -= 1

    open_br = [b for b in count.keys() if (count[b] >= max(count.values()) > 0) and (count[b] % 2 == 1) ]

    pos = [text.rfind( b ) for b in open_br]

    if pos:

        return max(pos) + 1

def function(text):

    tokens = text.strip()

    if tokens[-1] != ":":

        return

    tokens = tokens[:-1]
    
    for b in left_b + right_b:
        
        tokens = tokens.replace(b," ")

    tokens = tokens.split()

    for kw in py_indent_kw:

        if kw in tokens:

            return indented(text) + tabsize



"""
    FoxDot __main__.py
    ------------------

    Use FoxDot's interface by running this as a Python script, e.g.
    python __main__.py or python -m FoxDot if you have FoxDot correctly
    installed and Python on your path.

"""


from __future__ import absolute_import, division, print_function

from .lib import FoxDotCode, handle_stdin

import argparse

parser = argparse.ArgumentParser(
    prog="FoxDot", 
    description="Live coding with Python and SuperCollider", 
    epilog="More information: https://foxdot.org/")

parser.add_argument('-p', '--pipe', action='store_true', help="run FoxDot from the command line interface")
parser.add_argument('-d', '--dir', action='store', help="use an alternate directory for looking up samples")
parser.add_argument('-s', '--startup', action='store', help="use an alternate startup file")
parser.add_argument('-S', '--simple', action='store_true', help="run FoxDot in simple (accessible) mode")
parser.add_argument('-n', '--no-startup', action='store_true', help="does not load startup.py on boot")
parser.add_argument('-b', '--boot', action='store_true', help="Boot SuperCollider from the command line")

args = parser.parse_args()

if args.dir:

    try:

        # Use given directory

        FoxDotCode.use_sample_directory(args.dir)

    except OSError as e:

        # Exit with last error

        import sys, traceback
        sys.exit(traceback.print_exc(limit=1))

if args.startup:

    try:

        FoxDotCode.use_startup_file(args.startup)

    except OSError as e:

        import sys, traceback
        sys.exit(traceback.print_exc(limit=1))

if args.no_startup:

    FoxDotCode.no_startup()

if args.boot:

    FoxDotCode.boot_supercollider()

if args.pipe:

    # Just take commands from the CLI

    handle_stdin()

else:

    # Open the GUI

    if args.simple:

        from .lib.Workspace.Simple import workspace

    else:

        from .lib.Workspace.Editor import workspace

    FoxDot = workspace(FoxDotCode).run()



from __future__ import absolute_import, division, print_function

from .tkimport import *

from ..Settings import FOXDOT_ICON, FOXDOT_ICON_GIF

try:
    import tkMessageBox
except ImportError:
    from tkinter import messagebox as tkMessageBox

import os.path


class Config:
    def __init__(self, path):
        self.root = Toplevel()
        self.root.title("conf.txt")
        self.root.protocol("WM_DELETE_WINDOW", self.save_and_close )

        try:

            # Use .ico file by default
            self.root.iconbitmap(FOXDOT_ICON)
            
        except TclError:

            # Use .gif if necessary
            self.root.tk.call('wm', 'iconphoto', self.root._w, PhotoImage(file=FOXDOT_ICON_GIF))

        self.filepath = os.path.realpath(path)

        self.y_scroll = Scrollbar(self.root)
        self.y_scroll.grid(row=0, column=2, sticky='nsew')

        self.textbox = Text(self.root, width=50, yscrollcommand=self.y_scroll.set)
        self.textbox.grid(row=0, column=0, columnspan=2)
        self.y_scroll.config(command=self.textbox.yview)

        self.exit = Button(self.root, text="Cancel", command=self.save_and_close)
        self.exit.grid(row=1, column=0, stick=N+S+E+W)

        self.save = Button(self.root, text="Save Changes", command=self.save_changes)
        self.save.grid(row=1, column=1, sticky=N+S+E+W, columnspan=2)

        self.unsaved = True

        with open(self.filepath) as f:

            self.text = f.read().rstrip()

        self.textbox.insert(INSERT, self.text)

        # Add binds?

        # self.textbox.bind()
        
    def start(self):
        self.root.mainloop()

    def save_and_close(self, event=None):
        """ Asks the user if they want to save changes """
        if self.get_text() != self.text:
            answer = tkMessageBox.askyesno("Save changes", "Do you want to save your changes?")
            if answer:
                return self.save_changes()
        return self.root.destroy()

    def get_text(self):
        return self.textbox.get("0.0", END).rstrip()

    def save_changes(self):
        text = self.get_text()
        f = open(self.filepath, "w") # writing a file
        f.write(text)
        f.close()
        self.root.destroy()
        tkMessageBox.showwarning(title="Just a heads up", message="A restart of FoxDot is required for the changes to take effect")
        return




# Tutorial 11: Playing Custom Samples

# You can use your own samples by simply dropping audio files into the existing FoxDot sample directories.
# These are found in the 'snd' directory in the root of the FoxDot installation
# (e.g., 'C:\Python27\Lib\site-packages\FoxDot\snd').

# You saw earlier how to work with samples using play(). You can also play samples with loop().
s1 >> loop('foxdot')

# You may notice that this is just playing the first part of the sample over and over again.
# You can tweak the behavior with many of the arguments we've seen thus far for controlling other synths. dur is a good place to start.
s1 >> loop('foxdot', dur=4)

# If you have a folder full of samples that you would like to use in FoxDot, you can call loop() with the full path to the sample.
s1 >> loop('/path/to/samples/quack.wav')

# If you give loop the path to a folder, it will play the first sample it finds. You can change which sample it plays with the sample= arg.

# Play the first sample in my collection
s1 >> loop('/path/to/samples')

# Play the second sample in my collection
s1 >> loop('/path/to/samples', sample=1)

# If you're going to be using a lot of samples from a folder, you can add it to the sample search path. FoxDot will look under all its search paths for a matching sample when you give it a name.
Samples.addPath('/path/to/samples')
s1 >> loop('quack')

# Once you have a search path, you can use pattern matching to search for samples.

# Play the 3rd sample under the 'snare' dir
s1 >> loop('snare/*', sample=2)

# You can use * in directory names too
s1 >> loop('*_120bpm/drum*/kick*')

# ** means "all recursive subdirectories". This will play the first sample
# nested under 'percussion' (e.g. 'percussion/kicks/classic/808.wav')
s1 >> loop('percussion/**/*')



"""
    Filter Effects
    --------------

    Effects are added to Player objects as keywords instructions like `dur`
    or `amp` but are a little more tricky. Each effect has a "title" keyword,
    which requires a nonzero value to add the effect to a Player. Effects
    also have other attribute keywords which can be any value and may have
    a default value which is set when a Player is created.

    ::
        # Example. Reverb effect "title" is `room` and attribute is `mix`, which
        # defaults to 0.25. The following adds a reverb effect
        
        p1 >> pads(room=0.5)

        # This still adds the effect, but a mix of 0 doesn't actually do anything

        p1 >> pads(room=0.5, mix=0)

        # This effect is not added as the "title" keyword, room, is 0

        p1 >> pads(room=0, mix=0.5)

    Other effects are outlined below:
    
    *High Pass Filter* - Title keyword: `hpf`, Attribute keyword(s): `hpr`
    Only frequences **above** the value of `hpf` are kept in the final signal. Use `hpr` to set the resonance (usually a value between 0 and 1)

    *Low Pass Filter* - Title keyword: `lpf`, Attribute keyword(s): `lpr`
    Only frequences **below** the value of `lpf` are kept in final signal. Use `lpr` to set the resonance (usually a value between 0 and 1)

    *Bitcrush* - Title keyword: `bits`, Attribute keyword(s): `crush`
    The bit depth, in number of `bits`, that the signal is reduced to; this is a value between 1 and 24 where other values are ignored. Use `crush` to set the amount of reduction to the bitrate (defaults to 8)

    *Reverb* - Title keyword: `room`, Attribute keyword(s): `mix`
    The `room` argument specifies the size of the room and `mix` is the dry/wet mix of reverb; this should be a value between 0 and 1 (defalts to 0.25)

    *Chop* - Title keyword: `chop`, Attribute keyword(s): `sus`
    'Chops' the signal into chunks using a low frequency pulse wave over the sustain of a note.

    *Slide To* - Title keyword: `slide`, Attribute keyword(s):
    Slides' the frequency value of a signal to `freq * (slide+1)` over the  duration of a note (defaults to 0)

    *Slide From* - Title keyword: `slidefrom`, Attribute keyword(s):
    Slides' the frequency value of a signal from `freq * (slidefrom)` over the  duration of a note (defaults to 1)

    *Comb delay (echo)* - Title keyword: `echo`, Attribute keyword(s): `decay`
    Sets the decay time for any echo effect in beats, works best on Sample Player (defaults to 0)

    *Panning* - Title keyword: `pan`, Attribute keyword(s):
    Panning, where -1 is far left, 1 is far right (defaults to 0)

    *Vibrato* - Title keyword: `vib`, Attribute keyword(s): 
    Vibrato (defaults to 0)

    Undocumented: Spin, Shape, Formant, BandPassFilter, Echo


"""

from __future__ import absolute_import, division, print_function

import os.path
from ..Settings import EFFECTS_DIR, SC3_PLUGINS
from ..ServerManager import Server

class Effect:
    server=Server
    def __init__(self, foxdot_name, synthdef, args={}, control=False):

        self.name      = foxdot_name
        self.synthdef  = synthdef
        self.filename  = EFFECTS_DIR + "/{}.scd".format(self.synthdef)
        self.args      = args.keys()
        self.vars      = ["osc"]
        self.defaults  = args
        self.effects   = []
        self.control   = control

        self.suffix    = "kr" if self.control else "ar"
        self.channels  = 1 if self.control else 2

        self.input     = "osc = In.{}(bus, {});\n".format(self.suffix, self.channels)
        self.output    = "ReplaceOut.{}".format(self.suffix)
        
    @classmethod
    def set_server(cls, server):
        cls.server = server
    
    def __repr__(self):
        # return "<Fx '{}' -- args: {}>".format(self.synthdef, ", ".join(self.args))
        other_args = ['{}'.format(arg) for arg in self.args if arg != self.name]
        other_args = ", other args={}".format(other_args) if other_args else ""
        return "<'{}': keyword='{}'{}>".format(self.synthdef, self.name, other_args)

    def __str__(self):
        s  = "SynthDef.new(\{},\n".format(self.synthdef)
        s += "{" + "|bus, {}|\n".format(", ".join(self.args))
        s += "var {};\n".format(",".join(self.vars))
        s += self.input
        s += self.list_effects()
        s += self.output
        s += "(bus, osc)}).add;"
        return s

    def add(self, string):
        self.effects.append(string)
        return

    def doc(self, string):
        """ Set a docstring for the effects"""
        return 

    def list_effects(self):
        s = ""
        for p in self.effects:
            s += p + ";\n"
        return s

    def add_var(self, name):
        if name not in self.vars:
            self.vars.append(name)
        return

    def save(self):
        ''' writes to file and sends to server '''
        
        # 1. See if the file exists

        if os.path.isfile(self.filename):

            with open(self.filename) as f:

                contents = f.read()

        else:

            contents = ""

        # 2. If it does, check contents

        this_string = self.__str__()

        if contents != this_string:

            try:

                with open(self.filename, 'w') as f:
                
                    f.write(this_string)

            except IOError:

                print("IOError: Unable to update '{}' effect.".format(self.synthdef))

        # 3. Send to server

        self.load()

    def load(self):
        """ Load the Effect """
        if self.server is not None:
            self.server.loadSynthDef(self.filename)
        return

class In(Effect):
    def __init__(self):
        Effect.__init__(self, 'startSound', 'startSound')
        self.save()      
    def __str__(self):
        s  = "SynthDef.new(\startSound,\n"
        s += "{ arg bus, rate=1, sus; var osc;\n"
        s += "	ReplaceOut.kr(bus, rate)}).add;\n"
        return s

class Out(Effect):
    def __init__(self):
        self.max_duration = 8
        Effect.__init__(self, 'makeSound', 'makeSound')
        self.save()
    def __str__(self):
        s  = "SynthDef.new(\makeSound,\n"
        s += "{ arg bus, sus; var osc;\n"
        s += "	osc = In.ar(bus, 2);\n"
        s += "  osc = EnvGen.ar(Env([1,1,0],[sus * {}, 0.1]), doneAction: 14) * osc;\n".format(self.max_duration)
        s += "	DetectSilence.ar(osc, amp:0.0001, time: 0.1, doneAction: 14);\n"
        #s += "	Out.ar(0, osc);\n"
        s += "OffsetOut.ar(0, osc[0]);\n"
        s += "OffsetOut.ar(1, osc[1]);\n"
        s += " }).add;\n"
        return s

class EffectManager(dict):
    def __init__(self):

        dict.__init__(self)
        self.kw=[]
        self.all_kw=[]
        self.defaults={}
        self.order={N:[] for N in range(3)}

    def __repr__(self):
        return "\n".join([repr(value) for value in self.values()])

    def values(self):
        return [self[key] for key in self.sort_by("synthdef")]

    def sort_by(self, attr):
        """ Returns the keys sorted by attribute name"""
        return sorted(self.keys(), key=lambda effect: getattr(self[effect], attr))

    def new(self, foxdot_arg_name, synthdef, args, order=2):
        self[foxdot_arg_name] = Effect(foxdot_arg_name, synthdef, args, order==0)

        if order in self.order:

            self.order[order].append(foxdot_arg_name)

        else:

            self.order[order] = [foxdot_arg_name]

        # Store the main keywords together

        self.kw.append(foxdot_arg_name)

        # Store other sub-keys

        for arg in args:
            if arg not in self.all_kw:
                self.all_kw.append(arg)

            # Store the default value
            
            self.defaults[arg] = args[arg]

        return self[foxdot_arg_name]
    
    def kwargs(self):
        """ Returns the title keywords for each effect """
        return tuple(self.kw)

    def all_kwargs(self):
        """ Returns *all* keywords for all effects """
        return tuple(self.all_kw)

    def __iter__(self):
        for key in self.kw:
            yield key, self[key]

    def reload(self):
        """ Re-sends each effect to SC """
        for kw, effect in self:
            effect.load()
        In(); Out();
        return


# -- TODO

# Have ordered effects e.g.
# 0. Process frequency / playback rate
# 1. Before envelope
# 2. Adding the envelope
# 3. After envelope

FxList = EffectManager()

Effects = FxList # Alias - to become default

# Frequency Effects

fx = FxList.new("vib", "vibrato", {"vib": 0, "vibdepth": 0.02}, order=0)
fx.add("osc = Vibrato.ar(osc, vib, depth: vibdepth)")
fx.save()

fx = FxList.new("slide", "slideTo", {"slide":0, "sus":1, "slidedelay": 0}, order = 0)
fx.add("osc = osc * EnvGen.ar(Env([1, 1, slide + 1], [sus*slidedelay, sus*(1-slidedelay)]))")
fx.save()

fx = FxList.new("slidefrom", "slideFrom", {"slidefrom": 0, "sus": 1, "slidedelay": 0}, order=0)
fx.add("osc = osc * EnvGen.ar(Env([slidefrom + 1, slidefrom + 1, 1], [sus*slidedelay, sus*(1-slidedelay)]))")
fx.save()

fx = FxList.new("glide", "glissando", {"glide": 0, "glidedelay": 0.5, "sus": 1}, order=0)
fx.add("osc = osc * EnvGen.ar(Env([1, 1, (1.059463**glide)], [sus*glidedelay, sus*(1-glidedelay)]))")
fx.save()

fx = FxList.new("bend", "pitchBend", {"bend": 0, "sus": 1, "benddelay": 0}, order=0)
fx.add("osc = osc * EnvGen.ar(Env([1, 1, 1 + bend, 1], [sus * benddelay, (sus*(1-benddelay)/2), (sus*(1-benddelay)/2)]))")
fx.save()

fx = FxList.new("coarse", "coarse", {"coarse": 0, "sus": 1}, order=0)
fx.add("osc = osc * LFPulse.ar(coarse / sus)")
fx.save()

fx = FxList.new("striate", "striate", {"striate": 0, "sus": 1, "buf": 0, "rate": 1}, order=0)
fx.add("rate = (BufDur.kr(buf) / sus)")
fx.add("rate = Select.kr(rate > 1, [1, rate])")
fx.add("osc = osc * LFPulse.ar(striate / sus, width:  (BufDur.kr(buf) / rate) / sus) * rate")
fx.save()

fx = FxList.new("pshift", "pitchShift", {"pshift":0}, order=0)
fx.add("osc = osc * (1.059463**pshift)")
fx.save()

# fx = FxList.new("fm_sin", "FrequencyModulationSine", {"fm_sin":0, "fm_sin_i":1}, order=0)
# fx.add("osc = osc + (fm_sin_i * SinOsc.kr(osc * fm_sin))")
# fx.save()

# fx = FxList.new("fm_saw", "FrequencyModulationSaw", {"fm_saw":0, "fm_saw_i":1}, order=0)
# fx.add("osc = osc + (fm_saw_i * Saw.kr(osc * fm_saw))")
# fx.save()

# fx = FxList.new("fm_pulse", "FrequencyModulationPulse", {"fm_pulse":0, "fm_pulse_i":1}, order=0)
# fx.add("osc = osc + (fm_pulse_i * Pulse.kr(osc * fm_pulse))")
# fx.save()

# Signal effects

fx = FxList.new('hpf','highPassFilter', {'hpf': 0, 'hpr': 1}, order=2)
fx.doc("Highpass filter")
fx.add('osc = RHPF.ar(osc, hpf, hpr)')
fx.save()

fx = FxList.new('lpf','lowPassFilter', {'lpf': 0, 'lpr': 1}, order=2)
fx.add('osc = RLPF.ar(osc, lpf, lpr)')
fx.save()

fx = FxList.new('swell','filterSwell', {'swell': 0, 'sus': 1, 'hpr': 1}, order=2)
fx.add_var("env")
fx.add("env = EnvGen.kr(Env([0,1,0], times:[(sus*0.25), (sus*0.25)], curve:\\sin))")
fx.add('osc = RHPF.ar(osc, env * swell * 2000, hpr)')
fx.save()

fx = FxList.new("bpf", "bandPassFilter", {"bpf": 0, "bpr": 1, "bpnoise": 0, "sus": 1}, order=2)
fx.add("bpnoise = bpnoise / sus")
fx.add("bpf = LFNoise1.kr(bpnoise).exprange(bpf * 0.5, bpf * 2)")
fx.add("bpr = LFNoise1.kr(bpnoise).exprange(bpr * 0.5, bpr * 2)")
fx.add("osc = BPF.ar(osc, bpf, bpr)")
fx.save()
       
if SC3_PLUGINS:

    fx = FxList.new('crush', 'bitcrush', {'bits': 8, 'sus': 1, 'amp': 1, 'crush': 0}, order=1)
    fx.add("osc = Decimator.ar(osc, rate: 44100/crush, bits: bits)")
    fx.add("osc = osc * Line.ar(amp * 0.85, 0.0001, sus * 2)") 
    fx.save()

    fx = FxList.new('dist', 'distortion', {'dist': 0, 'tmp': 0}, order=1)
    fx.add("tmp = osc")
    fx.add("osc = CrossoverDistortion.ar(osc, amp:0.2, smooth:0.01)")
    fx.add("osc = osc + (0.1 * dist * DynKlank.ar(`[[60,61,240,3000 + SinOsc.ar(62,mul:100)],nil,[0.1, 0.1, 0.05, 0.01]], osc))")
    fx.add("osc = (osc.cubed * 8).softclip * 0.5")
    fx.add("osc = SelectX.ar(dist, [tmp, osc])")
    fx.save()

# Envelope -- just include in the SynthDef and use asdr?

# Post envelope effects    

fx = FxList.new('chop', 'chop', {'chop': 0, 'sus': 1}, order=2)
fx.add("osc = osc * LFPulse.kr(chop / sus, add: 0.01)")
fx.save()

fx = FxList.new('tremolo', 'tremolo', {'tremolo': 0, 'beat_dur': 1}, order=2)
fx.add("osc = osc * SinOsc.ar( tremolo / beat_dur, mul:0.5, add:0.5)")
fx.save()

fx = FxList.new('echo', 'combDelay', {'echo': 0, 'beat_dur': 1, 'echotime': 1}, order=2)
fx.add('osc = osc + CombL.ar(osc, delaytime: echo * beat_dur, maxdelaytime: 2 * beat_dur, decaytime: echotime * beat_dur)')
fx.save()

fx = FxList.new('spin', 'spinPan', {'spin': 0,'sus': 1}, order=2)
fx.add('osc = osc * [FSinOsc.ar(spin / 2, iphase: 1, mul: 0.5, add: 0.5), FSinOsc.ar(spin / 2, iphase: 3, mul: 0.5, add: 0.5)]')
fx.save()

fx = FxList.new("cut", "trimLength", {"cut": 0, "sus": 1}, order=2)
fx.add("osc = osc * EnvGen.ar(Env(levels: [1,1,0.01], curve: 'step', times: [sus * cut, 0.01]))")
fx.save()

fx = FxList.new('room', 'reverb', {'room': 0, 'mix': 0.1}, order=2)
fx.add("osc = FreeVerb.ar(osc, mix, room)")
fx.save()

fx = FxList.new("formant", "formantFilter", {"formant": 0}, order=2)
fx.add("formant = (formant % 8) + 1")
fx.add("osc = Formlet.ar(osc, formant * 200, ((formant % 5 + 1)) / 1000, (formant * 1.5) / 600).tanh")
fx.save()

fx = FxList.new("shape", "wavesShapeDistortion", {"shape":0}, order=2)
fx.add("osc = (osc * (shape * 50)).fold2(1).distort / 5")
fx.save()

fx = FxList.new("drive", "overdriveDistortion", {"drive":0}, order=2)
fx.add("osc = (osc * (drive * 50)).clip(0,0.2).fold2(2)")
fx.save()

In(); Out()
Effect.server.setFx(FxList)



import wx.stc
import wx

class Text(wx.TextCtrl):
    """docstring for SimpleText"""
    def __init__(self, *args, **kwargs):
        wx.TextCtrl.__init__(self, *args, **kwargs)
        self._default_style   = wx.TextAttr(wx.Colour(0, 0, 0), wx.Colour(255,255,255))
        self._highlight_style = wx.TextAttr(wx.Colour(255,255,255), wx.Colour(255,0,0))

    # Return values

    def get_insert(self):
        _, col, row = self.PositionToXY(self.GetInsertionPoint())
        return (row, col)

    def get_current_row(self):
        return self.get_insert()[0]

    def get_current_block(self):
        """ Returns the start and end rows of the current block """
        anchor, _ = self.get_insert()

        # Go forward
        row = anchor
        while True:
            text = self.GetLineText(row).strip()
            if len(text) == 0:
                end = row - 1
                break
            row += 1

        # Go backwards
        row = anchor
        while True:
            text = self.GetLineText(row).strip()
            if len(text) == 0:
                start = row + 1
                break
            row -= 1            
            
        return (start, end)

    def get_current_line_text(self):
        """ Returns the text from the current line """
        return self.GetLineText(self.get_insert()[0])

    def get_block_text(self, start, end):
        """ Returns the text between """
        return "\n".join([self.GetLineText(row) for row in range(start, end + 1)])

    # Highlight block

    def highlight_block(self, start, end, undo=False):
        for row in range(start, end + 1):
            if undo:
                self.de_highlight_line(row)
            else:
                self.highlight_line(row)
        return

    def highlight_line(self, row):
        start = self.XYToPosition(0, row)
        end   = self.XYToPosition(self.GetLineLength(row), row)
        self.SetStyle(start, end, self._highlight_style)
        return

    def de_highlight_line(self, row):
        start = self.XYToPosition(0, row)
        end   = self.XYToPosition(self.GetLineLength(row), row)
        val = self.SetStyle(start, end, self._default_style)
        return

    def delete_all(self):
        """ Removes all the text in the editor """
        return self.Remove(0, self.GetLastPosition())



"""
    compare_functions.py(funcA, funcB):

        returns True if the functions are identical
    
"""

from __future__ import absolute_import, division, print_function

def func_cmp(funcA, funcB):

    codeA = funcA.__code__
    A_bytecode  = codeA.co_code
    A_constants = codeA.co_consts
    A_variables = codeA.co_names

    codeB = funcB.__code__
    B_bytecode  = codeB.co_code
    B_constants = codeB.co_consts
    B_variables = codeB.co_names
    
    return all([A_bytecode == B_bytecode,
                A_constants == B_constants,
                A_variables == B_variables])

def func_str(func):
    """ Returns a function as a string """
    code = func.__code__
    return ",".join([func.__name__, str(code.co_names), str(code.co_consts)])



"""
Time-Dependent Variables (TimeVar)

TODO: using __call__ -> go through getattribute and check instead of already having a __call__

"""

from __future__ import absolute_import, division, print_function

from .Patterns import *
from .Utils  import *
from .Patterns.Operations import *
from .Constants import inf 

from time import time

def fetch(func):
    """ Function to wrap basic lambda operators for TimeVars  """
    def eval_now(a, b):
        if isinstance(a, TimeVar):
            a = a.now()
        if isinstance(b, TimeVar):
            b = b.now()
        return func(a, b)
    return eval_now


class TimeVar(object):
    """ Var(values [,durs=[4]]) """

    metro = None
    depth = 128

    def __init__(self, values, dur=None, start=0, **kwargs):

        if dur is None:

            dur = self.metro.bar_length()

        self.name     = "un-named"

        self.start_time = float(start) # offset

        self.values   = values
        self.dur      = dur
        self.bpm      = kwargs.get('bpm', None)

        self.get_seconds = bool(kwargs.get('seconds', False))

        # Dynamic method for calculating values
        self.func     = Nil
        self.evaluate = fetch(Nil)
        self.dependency = None

        self.update(values, dur)

        self.current_value = None
        self.current_index = 0
        self.next_value    = None
        self.next_time     = 0
        self.prev_time     = 0
        self.next_index    = 0

        # Private flags

        self.__accessed = False
        self.__inf_index = None

        self.proportion    = 0

        # If the clock is not ticking, start it

        if self.metro.ticking == False:

            self.metro.start()

    def json_value(self):
        """ Returns data about this TimeVar that can be sent over a network as JSON  """
        ## pickle?
        return [str(self.__class__.__name__), list(self.values), list(self.dur)]

    @classmethod
    def set_clock(cls, tempo_clock):
        cls.metro = tempo_clock
        return

    @classmethod
    def help(cls):
        return print(cls.__doc__)

    @staticmethod
    def stream(values):
        return asStream(values)

    @staticmethod
    def CreatePvarGenerator(func, *args, **kwargs):
        return PvarGenerator(func, *args, **kwargs)

    # Standard dunder methods
    def __str__(self):
        return str(self.now())
    def __repr__(self):
        return str(self.now())
    def __len__(self):
        return len(self.now())
    def __int__(self):
        return int(self.now())
    def __float__(self):
        return float(self.now())
    def __abs__(self):
        return abs(self.now())

    # For printing the details
    def info(self):
        return "<{}({}, {})>".format(self.__class__.__name__, repr(self.get_values()), repr(self.get_durs()))

    def all_values(self):
        """ Displays the values and the dependency value - useful for debugging """
        return self.value + [self.dependency]

    # Update methods

    def new(self, other):
        """ Returns a new TimeVar object """
        # new = TimeVar(other, self.dur, bpm=self.bpm)
        new = ChildTimeVar(other)
        new.dependency = self
        return new

    def update(self, values, dur=None, **kwargs):
        """ Updates the TimeVar with new values.
        """

        self.bpm = kwargs.get('bpm', self.bpm)

        #: Update the durations of each state

        if dur is not None:

            self.dur = asStream(dur)

        self.values = self.stream(values)

        return self

    def get_current_index(self, time=None):
        """ Returns the index of the value currently represented """

        # Get the time value if not from the Clock

        time = self.get_current_time(time) - self.start_time

        if self.get_inf_index() is not None:

            return self.get_inf_index()

        if time >= self.next_time:

            while True:

                next_dur = self.dur[self.next_index]

                self.next_time, self.prev_time = self.next_time + next_dur, self.next_time

                # If we find an "inf"

                if self.check_for_inf(next_dur):

                    self.set_inf_index(self.next_index)

                    return self.get_inf_index()

                self.next_index += 1

                if self.next_time >= time:

                    break

        # Store the % way through this value's time

        try:

            self.proportion = float((time - self.prev_time) / (self.next_time - self.prev_time))

        except ZeroDivisionError:

            self.proportion = 1.0

        # The current index is the next index minus one

        self.current_index = self.next_index - 1

        # Flag we have accessed the value

        self.flag_accessed()

        return self.current_index

    # Inf

    def set_inf_index(self, value):
        self.__inf_index = int(value)
        return

    def get_inf_index(self):
        self.proportion = 0
        return self.__inf_index

    def check_for_inf(self, duration):
        return (self.__accessed and duration == inf)

    def flag_accessed(self):
        self.__accessed = True
        return

    # Evaluation methods

    def calculate(self, val): # maybe rename to resolve
        """ Returns val as modified by its dependencies """
        return self.evaluate(val, self.dependency)

    def get_current_time(self, beat=None):
        """ Returns the current beat value """
        # Return elapsed time in seconds if get_seconds flag is True
        if self.get_seconds is True:
            return float(self.metro.time)
        # Else return the beat
        if beat is None:
            beat = self.metro.now()
        if self.bpm is not None:
            beat *= (self.bpm / float(self.metro.bpm))
        return float(beat)

    def now(self, time=None):
        """ Returns the value currently represented by this TimeVar """

        i = self.get_current_index(time)
        self.current_value = self.calculate(self.values[i])
        
        return self.current_value

    def copy(self):
        new = var(self.values, self.dur, bpm=self.bpm)
        return new

    def get_durs(self):
        return self.dur

    def get_values(self):
        return self.values

    # 1. Methods that change the 'var' in place
    def i_invert(self):
        lrg = float(max(self.values))
        for i, item in enumerate(self.values):
            self.values[i] = (((item / lrg) * -1) + 1) * lrg
        return

    # Method that return an augmented NEW version of the 'var'

    def invert(self):
        new = self.new(self.values)
        lrg = float(max(new.data))
        for i, item in enumerate(new.data):
            new.data[i] = (((item / lrg) * -1) + 1) * lrg
        return new

    def lshift(self, duration):
        time = [self.dur[0]-duration] + list(self.dur[1:]) + [duration]
        return self.__class__(self.values, time)

    def rshift(self, duration):
        time = [duration] + list(self.dur[:-1]) + [self.dur[-1]-duration]
        data = [self.values[-1]] + list(self.values)
        return self.__class__(data, time)

    def extend(self, values, dur=None):
        data = list(self.values) + list(values)
        durs = self.dur if not dur else list(self.dur) + list(asStream(dur))
        return self.__class__(data, durs)

    def shuf(self):
        pass

    # Mathmetical operators

    def math_op(self, other, op):
        """ Performs the mathematical operation between self and other. "op" should 
            be the string name of a dunder method  e.g. __mul__ """
        if not isinstance(other, (TimeVar, int, float)):
            if type(other) is tuple:
                return PGroup([getattr(self, op).__call__(x) for x in other])
            elif type(other) is list:
                return Pattern([getattr(self, op).__call__(x) for x in other])
            else:
                return getattr(other, get_inverse_op(op)).__call__(self)
        return other

    def set_eval(self, func):
        self.evaluate = fetch(func)
        self.func     = func
        return

    def __add__(self, other):
        new = self.math_op(other, "__add__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(Add)
        return new

    def __radd__(self, other):
        new = self.math_op(other, "__radd__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(rAdd)
        return new

    def __sub__(self, other):
        new = self.math_op(other, "__sub__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(rSub)
        return new

    def __rsub__(self, other):
        new = self.math_op(other, "__rsub__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(Sub)
        return new

    def __mul__(self, other):
        new = self.math_op(other, "__mul__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(Mul)
        return new

    def __rmul__(self, other):
        new = self.math_op(other, "__rmul__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(Mul)
        return new

    def __pow__(self, other):
        new = self.math_op(other, "__pow__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(rPow)
        return new

    def __rpow__(self, other):
        new = self.math_op(other, "__rpow__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(Pow)
        return new

    def __floordiv__(self, other):
        new = self.math_op(other, "__floordiv__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(rFloorDiv)
        return new

    def __rfloordiv__(self, other):
        new = self.math_op(other, "__rfloordiv__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(FloorDiv)
        return new

    def __truediv__(self, other):
        new = self.math_op(other, "__truediv__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(rDiv)
        return new

    def __rtruediv__(self, other):
        new = self.math_op(other, "__rtruediv__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(Div)
        return new

    # Incremental operators (use in place of var = var + n)
    def __iadd__(self, other):
        self.values = self.values + other
        return self
    def __isub__(self, other):
        self.values = self.values - other
        return self
    def __imul__(self, other):
        self.values = self.values * other
        return self
    def __idiv__(self, other):
        self.values = self.values / other
        return self

    # Comparisons -- todo: return TimeVars

    def __gt__(self, other):
        return float(self.now()) > float(other)

    def __lt__(self, other):
        return float(self.now()) < float(other)

    def __ge__(self, other):
        return float(self.now()) >= float(other)

    def __le__(self, other):
        return float(self.now()) >= float(other)

    # %
    def __mod__(self, other):
        new = self.math_op(other, "__mod__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(rMod)
        return new

    def __rmod__(self, other):
        new = self.math_op(other, "__rmod__")
        if not isinstance(other, (TimeVar, int, float)):
            return new
        new = self.new(other)
        new.evaluate = fetch(Mod)
        return new

    #  Comparisons -- todo: return TimeVar

    def __eq__(self, other):
        return other == self.now()

    def __ne__(self, other):
        return other != self.now()

    # Storing functions etc

    # def __call__(self, *args, **kwargs):
    #     """ A TimeVar can store functions and will call the current item with this method """
    #     if callable(self.now()):
    #         return self.now().__call__(*args, **kwargs)
    #     else:
    #         return self.now()

    # Emulating container types

    def __getitem__(self, other):
        new = self.new(other)
        new.dependency = self
        new.evaluate = fetch(rGet)
        return new

    def __iter__(self):
        for item in self.now():
            yield item

    def transform(self, func):
        """ Returns a new TimeVar based on a func """
        new = self.new(0)
        new.dependency = self
        new.evaluate = fetch(lambda a, b: func(b))
        return new

class ChildTimeVar(TimeVar):
    """ When a new TimeVar is created using a function such as addition,
        e.g. var([0,2]) + 2, then a ChildTimeVar is created that contains a
        single value but also creates a new ChildTimeVar when operated upon
        and behaves just as a TimeVar does."""
    def now(self, time=None):
        self.current_value = self.calculate(self.values[0])
        return self.current_value

class linvar(TimeVar):
    def now(self, time=None):
        """ Returns the value currently represented by this TimeVar """
        i = self.get_current_index(time)
        self.current_value = self.calculate(self.values[i])
        self.next_value    = self.calculate(self.values[i + 1])
        return self.get_timevar_value()

    def get_timevar_value(self):
        return (self.current_value * (1-self.proportion)) + (self.next_value * self.proportion)

class expvar(linvar):
    def get_timevar_value(self):
        self.proportion *= self.proportion
        return (self.current_value * (1-self.proportion)) + (self.next_value * self.proportion)

import math

class sinvar(linvar):
    def get_timevar_value(self):
        d = self.current_value  > self.next_value
        x = (self.proportion * 90) + (d * 270)
        self.proportion = math.sin(math.radians(x)) + int(d)
        return (self.current_value * (1-self.proportion)) + (self.next_value * self.proportion)

PATTERN_METHODS = Pattern.get_methods()

class Pvar(TimeVar):
    """ A TimeVar that represents Patterns that change over time e.g.
        ::

            >>> a = Pvar([ [0,1,2,3], [4,5] ], 4)
            >>> print a # time is 0
            P[0, 1, 2, 3]
            >>> print a # time is 4
            P[4, 5]
    """
    stream = PatternContainer
    def __init__(self, values, dur=None, **kwargs):

        try:

            data = [asStream(val) for val in values]

        except:

            data = [values]

        TimeVar.__init__(self, data, dur, **kwargs)


    def __get_pattern_attr(self, attr):
        """ Returns a function that transforms the patterns of this Pvar if the attr
            is a Pattern method, if not it returns the attribute  for the current pattern
        """

        pattern_attr = getattr(self.now(), attr)

        if callable(pattern_attr):

            def get_new_pvar(*args, **kwargs):

                # If this is the root Pvar, change the values

                if self.dependency is None:

                    print(len(self.values))

                    new_values = [getattr(pat, attr)(*args, **kwargs) for pat in self.values]

                    return Pvar(new_values, dur=self.dur)

                else:

                    # Get the "parent" Pvar and re-apply the connecting function

                    new_pvar = getattr(self.dependency, attr)(*args, **kwargs)

                    new_item = self.func(new_pvar, self.original_value)

                    return new_item

            return get_new_pvar

        else:

            return pattern_attr

    def getitem(self, index, *args, **kwargs):
        """ Returns a TimeVar based on getting the index of this Pattern """
        return TimeVar(0).transform(lambda e: self.now().getitem(index))

    def __getattr__(self, attr):
        """ (Python 2 compatability) Override for accessing pattern methods. Returns a new
            Pvar that has been "transformed" using the method such that then method also
            applies when values have been updated.  """

        try:

            return object.__getattr__(self, attr)

        except AttributeError:

            return self.__get_pattern_attr(attr)

    def __getattribute__(self, attr):
        """ Override for accessing pattern methods. Returns a new
            Pvar that has been "transformed" using the method such that then method also
            applies when values have been updated.  """

        try:

            return object.__getattribute__(self, attr)

        except AttributeError:

            return self.__get_pattern_attr(attr)

    def new(self, other):
        # new = Pvar([other], dur=self.dur)
        new = ChildPvar(other)
        new.original_value = other
        new.dependency = self
        return new

    def __getitem__(self, other):
        """ Return a single timevar when using getitem """
        new = ChildTimeVar(other)
        new.dependency = self
        new.evaluate = fetch(rGet)
        return new

    def set_eval(self, func):
        self.evaluate = fetch(func)
        self.func     = func
        return

    def __add__(self, other):
        new = self.new(other)
        new.set_eval(rAdd)
        return new

    def __radd__(self, other):
        new = self.new((other))
        new.set_eval(Add)
        return new

    def __sub__(self, other):
        new = self.new((other))
        new.set_eval(rSub)
        return new

    def __rsub__(self, other):
        new = self.new((other))
        new.set_eval(Sub)
        return new

    def __mul__(self, other):
        new = self.new((other))
        new.set_eval(rMul)
        return new

    def __rmul__(self, other):
        new = self.new((other))
        new.set_eval(Mul)
        return new

    def __div__(self, other):
        new = self.new((other))
        new.set_eval(rDiv)
        return new

    def __rdiv__(self, other):
        new = self.new((other))
        new.set_eval(Div)
        return new

    def __truediv__(self, other):
        new = self.new((other))
        new.set_eval(rDiv)
        return new

    def __rtruediv__(self, other):
        new = self.new((other))
        new.set_eval(Div)
        return new

    def __floordiv__(self, other):
        new = self.new((other))
        new.set_eval(rFloorDiv)
        return new

    def __rfloordiv__(self, other):
        new = self.new((other))
        new.set_eval(FloorDiv)
        return new

    def __pow__(self, other):
        new = self.new((other))
        new.set_eval(rPow)
        return new

    def __rpow__(self, other):
        new = self.new((other))
        new.set_eval(Pow)
        return new

    def __mod__(self, other):
        new = self.new((other))
        new.set_eval(rMod)
        return new

    def __rmod__(self, other):
        new = self.new((other))
        new.set_eval(Mod)
        return new

    def __or__(self, other):
        # Used when piping patterns together
        new = self.new(PatternContainer(other))
        new.set_eval(rOr)
        return new

    def __ror__(self, other):
        # Used when piping patterns together
        new = self.new(PatternContainer(other))
        new.set_eval(Or)
        return new

    def transform(self, func):
        """ Returns a Pvar based on a transformation function, as opposed to
            a mathematical operation"""
        new = self.new(self)
        new.set_eval(lambda a, b: b.transform(func))
        return new

class ChildPvar(Pvar):
    def now(self, time=None):
        self.current_value = self.calculate(self.values[0])
        return self.current_value


class PvarGenerator(Pvar):
    """ If a TimeVar is used in a Pattern function e.g. `PDur(var([3,5]), 8)`
        then a `PvarGenerator` is returned. Each argument is stored as a TimeVar
        and the function is called whenever the arguments are changed
    """
    def __init__(self, func, *args, **kwargs):
        self.p_func = func # p_func is the Pattern function e.g. PDur but self.func is created when operating on this PvarGenerator
        
        self.args = []

        if "pattern" in kwargs:

            self.args.append(kwargs["pattern"])
        
        self.args.extend( [(arg if isinstance(arg, TimeVar) else TimeVar(arg)) for arg in args] )

        self.last_args = []
        self.last_data = []
        self.evaluate = fetch(Nil)
        self.dependency = None

    def info(self):
        return "<{} {}>".format(self.__class__.__name__, self.func.__name__ + str(tuple(self.args)))

    def now(self):
        new_args = [arg.now() if isinstance(arg, TimeVar) else arg for arg in self.args]
        if new_args != self.last_args:
            self.last_args = new_args
            self.last_data = self.p_func(*self.last_args)
        pat = self.calculate(self.last_data)
        return pat

    def new(self, other):
        # new = Pvar([other]) # TODO -- test this
        new = self.__class__(lambda x: x, other)
        new.original_value = other
        new.dependency = self
        return new

    def set_eval(self, func):
        self.evaluate = fetch(func)
        self.func     = func
        return

    def __getattribute__(self, attr):
        # If it's a method, only return the method if its new, transform, or a dunder
        if attr in Pattern.get_methods():   
            
            if attr not in ("new", "now", "transform") and not attr.startswith("__"):

                # return a function that transforms the patterns of the  root Pvar

                def get_new_pvar_gen(*args, **kwargs):

                    # If this is the root Pvar, change the values

                    if self.dependency is None:

                        # Create a new function that combines the original *plus* the method

                        def new_func(*old_args, **old_kwargs):

                            return getattr(self.p_func(*old_args, **old_kwargs), attr)(*args, **kwargs)

                        return PvarGenerator(new_func, *self.args)

                    else:

                        # Get the "parent" Pvar and re-apply the connecting function

                        new_pvar_gen = getattr(self.dependency, attr)(*args, **kwargs)

                        return self.func(new_pvar_gen, self.original_value)

                return get_new_pvar_gen
                
        return object.__getattribute__(self, attr)

class PvarGeneratorEx(PvarGenerator):
    """ Un-Documented """
    def __init__(self, func, *args):
        self.func = func
        self.args = list(args)
        self.last_args = []
        self.last_data = []
        self.evaluate = fetch(Nil)
        self.dependency = 1

class mapvar(Pvar):
    """ Like a `Pvar`, the `mapvar` returns a whole `Pattern` as opposed to a single
        value, but instead of using the global clock to find the current value it
        uses the value in an instance of the `PlayerKey` class or another `TimeVar`. """
    def __init__(self, key, mapping, default=0):
        TimeVar.__init__(self, [])
        self.key     = key
        self.values  = {key: asStream(value) for key, value in mapping.items()}
        self.default = asStream(default)

    def get_current_index(self, time=None):
        self.current_index = self.key.now()
        return self.current_index

    def now(self, time=None):
        """ Returns the value currently represented by this TimeVar """
        i = self.get_current_index(time)
        self.current_value = self.calculate(self.values.get(i, self.default))
        return self.current_value

# Store and updates TimeVars

class _var_dict(object):
    """
        This is the TimeVar generator used in FoxDot. Calling it like `var()`
        returns a TimeVar but setting an attribute `var.foo = var([1,2],4)` will
        update the TimeVar that is already in `var.foo`.

        In short, using `var.name = var([i, j])` means you don't have to delete
        some of the text and replace it with `var.name.update([k, l])` you can
        just use `var.name = var([k, l])` and the contents of the var will be
        updated everywhere else in the program.
    """

    def __init__(self):
        self.__vars = {}
    @staticmethod
    def __call__(*args, **kwargs):
        return TimeVar(*args, **kwargs)
    def __setattr__(self, name, value):
        if name != "__vars" and isinstance(value, TimeVar):
            if name in self.__vars:
                if value.__class__ != self.__vars[name].__class__:
                    self.__vars[name].__class__ = value.__class__
                self.__vars[name].__dict__ = value.__dict__
            else:
                self.__vars[name] = value
            return
        object.__setattr__(self, name, value)
    def __getattr__(self, name):
        if name in self.__vars:
            value = self.__vars[name]
        else:
            try:
                value = object.__getattr__(self, name)
            except AttributeError:
                err = NameError("'var.{}' does not exist.".format(name))
                raise err
        return value

var = _var_dict()

# Give Main.Pattern a reference to TimeVar classes
Pattern.TimeVar       = TimeVar
Pattern.PvarGenerator = PvarGenerator
Pattern.Pvar          = Pvar



"""
    Patterns.Parse.py
    =================

    Handles the parsing of Sample Player Object Strings
"""

from __future__ import absolute_import, division, print_function

import re

from .PlayString import *
from .Generators import PRand
from .PGroups    import PGroupMod, PGroupOr, PGroupStar, PGroupPlus
from .Main       import Pattern, metaPattern, PatternMethod, PGroup, GeneratorPattern

from ..Utils import modi, LCM

re_nests  = r"\((.*?)\)"
re_square = r"\[.*?\]"
re_curly  = r"\{.*?\}"
re_arrow  = r"<.*?>"
square_type = PGroupPlus
braces_type = PRand
bar_type    = PGroupOr

def ParsePlayString(string, flat=False):
    """ Returns the parsed play string used by sample player """
    output, _ = feed(string)
    return output

def convert_to_int(data):
    """ Recursively calls until all nested data contains only integers """
    if isinstance(data, (int, float, str)):
        return int(data)
    elif isinstance(data, (list, tuple)):
        return data.__class__([convert_to_int(item) for item in data])
    elif isinstance(data, GeneratorPattern):
        return data.transform(convert_to_int)
    elif isinstance(data, metaPattern):
        return data.convert_data(convert_to_int)
    return int(data)

def arrow_zip(pat1, pat2):
    """ Zips two patterns together. If one item is a tuple, it extends the tuple / PGroup
        i.e. arrow_zip([(0,1),3], [2]) -> [(0,1,2),(3,2)]
    """
    output = Pattern()

    for i in range(LCM(len(pat1), len(pat2))):

        item1 = pat1.getitem(i, get_generator=True)
        item2 = pat2.getitem(i, get_generator=True)

        if all([x.__class__== PGroup for x in (item1, item2)]):

            new_item = PGroup(item1.data + item2.data)

        elif item1.__class__ == PGroup:

            new_item = PGroup(item1.data + [item2])

        elif item2.__class__ == PGroup:

            new_item = PGroup([item1] + item2.data)

        else:

            new_item = (item1, item2)

        output.append(new_item)

    return output

def feed(string):
    """ Used to recursively parse nested strings, returns a list object (not Pattern),
        and a boolean denoting if the list contains a nested list """
    
    string = PlayString(string)
    items  = [] # The actual pattern

    layer_pattern = False
    contains_nest = False
    
    i = 0
    
    while i < len(string):

        char = string[i]

        # look for a '<>'
        if char == "<":

            # Parse the contents of the brackets if found
            j = string.index(">", start=i+1)
            s = string[i+1:j]
            i = j

            chars, _ = feed(s)

            if len(chars) == 0:

                e = "Empty '<>' brackets in string"

                raise ParseError(e)

            # If we know we are layering, zip the last item

            if layer_pattern:

                items[-1] = items[-1].zip( Pattern(chars) )

            else:

                items.append(Pattern(chars))

                layer_pattern = True

            contains_nest = True

        # Look for || for specifying sample numbers

        elif char == "|":

            # Parse the contents of the brackets if found
            j = string.next_char_index("|", start=i+1)
            s = string[i+1:j]
            i = j

            chars, _ = feed(s)

            if len(chars) == 0:

                e = "Empty '||' delimeters in string"

                raise ParseError(e)

            try:

                assert(len(chars) == 2)

            except AssertionError:

                e = "'||' delimeters must contain exactly 2 elements"

                raise ParseError(e)

            # First is our list of sample chars
            
            samp_chr = chars[0]

            # Next is a list of integers for sample kw

            samp_num = convert_to_int(chars[1])

            # print(samp_chr, samp_num)

            items.append(bar_type((samp_chr, samp_num)))

        # Look for a '()'
        elif char == "(":

            # Parse the contents of the brackets if found
            j = string.index(")", start=i+1)
            
            s = string[i+1:j]
            i = j

            chars, _ = feed(s)

            if len(chars) == 0:

                e = "Empty '()' brackets in string"

                raise ParseError(e)

            items.append( chars ) # add the nested list

            layer_pattern = False

            contains_nest = True

        # Look for a '{}'
        elif char == "{":

            # Parse the contents of the brackets if found
            j = string.index("}", start=i+1)
            s = string[i+1:j]
            i = j

            chars, _ = feed(s)

            if len(chars) == 0:

                e = "Empty '{}' brackets in string"

                raise ParseError(e)

            items.append( braces_type(chars) )

            layer_pattern = False
                
        # Look for a '[]'
        elif char == "[":
            
            j = string.index("]", start=i+1)
            s = string[i+1:j]
            i = j

            chars, contains_nest = feed(s)

            if len(chars) == 0:

                e = "Empty '[]' brackets in string"

                raise ParseError(e)

            # Un-nest
            if contains_nest:

                # May contain sub-nests, so re-parse with calculated duration

                new_chars = []

                largest_item = max([len(ch) for ch in chars])

                for num in range(largest_item):

                    new_chars.append(square_type([modi(ch, num) for ch in chars]))

                items.append( new_chars )

                layer_pattern = False

            else:

                new_chars = []

                for char in chars:

                    new_chars.append(char)

                items.append( square_type(new_chars) )

                layer_pattern = False

        # Add single character to list

        elif char not in ")]}>|":

            items.append( char )

            layer_pattern = False

        # Increase iterator
        i += 1

    return items, contains_nest


@PatternMethod
def fromString(self, string):
    self.data = ParsePlayString(string)
    self.make()
    return self



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import re
from os.path import abspath, join, dirname

from ..Code import classes, functions, instances
from ..Settings import *

def path(fn):
    return abspath(join(dirname(__file__), fn))

# RegEx Group 1
user_defn = r"(?<=def )(\w+)"
players   = r"(?<=>>)(\s*\w+)"
comments  = r"^\s*#.*|[^\"']*(#[^\"']*$)"
decorator = r"\A\s*@.+"
#numbers   = r"(?<![a-zA-Z])\d+"
numbers   = r"\W+(\d+)"
strings   = r"\".*?\"|\".*" + "|\'.*?\'|\'.*"
dollar    = r"\s\$\s?"
arrow     = r"\s>>\s?"
multiline = r"\"\"\".*\"\"\"|'''.*'''"


# RegEx Group 2
def re_list(string, br="[]"):
    new = ""
    nested = False
    n, m = 0, 0
    while n < len(string):
        char = string[n]
        if char == br[0]:
            if nested is False:
                nested = True
                m = n
                n += 1
            else:
                sub = string[n + 1:string.rfind(br[1])]
                value = re_list(sub)
                n += len(value) + 1
        elif char == br[1]:
            new_string = string[m:n+1]
            return new_string
        else:
            n += 1

def re_pat(string):
    r = r"P?" + re_list(string)
    for char in "[]()":
        r = r.replace(char,"\\" + char)
    return re.search(r, string)

def findstyles(line, *args):
    """ Finds any locations of any regex and returns the name
        of the style and the start and end point in the line """

    if not args:

        tags = re_patterns

    else:

        tags = [a for a in args if a in re_patterns]

    pos = []

    for tag in tags:

        match_start = match_end = 0

        for match in re.finditer(re_patterns[tag], line):

            looking = True

            i = 0

            while looking:

                try:

                    start  = match.start(i)
                    end    = match.end(i)

                    if start == end == -1:

                        raise IndexError # this is hacky af

                    match_start = start
                    match_end   = end

                except IndexError:

                    looking = False

                i += 1

            pos.append((tag, match_start, match_end))

    return pos

def userdefined(line):
    """ Returns the name of a user defined function or class in a string
        i.e. def foo(x) -> foo """

    match = re.search(user_defn, line)

    if match:

        return match.group()

def find_comment(line):
    """ Finds the index of a comment # and returns None if not found """
    instring, instring_char = False, ""
    for i, char in enumerate(line):
        if char in ('"', "'"):
            if instring:
                if char == instring_char:
                    instring = False
                    instring_char = ""
            else:
                instring = True
                instring_char = char
        elif char == "#":
          if not instring:
              return i
    return None

def find_multiline(text):
    pos = []

    tag = multiline

    match_start = match_end = 0

    for match in re.finditer(tag, text, re.DOTALL):

        looking = True

        i = 0

        while looking:

            try:

                start  = match.start(i)
                end    = match.end(i)

                if start == end == -1:

                    raise IndexError # this is hacky af

                match_start = start
                match_end   = end

            except IndexError:

                looking = False

            i += 1

        pos.append((match_start, match_end))

    return pos


# Use our regex to read patterns.py and add all the functions to key_types
    
from ..Patterns import Main, Sequences, Generators
from ..SCLang import SCLang

foxdot_kw = [ "Clock","Group","Scale","Server","Root","Samples","var","Pvar",
              "Attributes", "PatternMethods", "PatternTypes", "now",
              "linvar","expvar","mapvar","sinvar","inf","when", "lambda", u"λ", decorator]

foxdot_funcs = classes(Main) + classes(Generators) + functions(Sequences) + ["P"]

# Python keywords used in RegEx Group 2

py_indent_kw = ["for","if","elif","else","def","while","class","try","except"]

py_functions = ["if","elif","else","return","def","print","when","from",
                 "and","or","not","is","in","for","as","with", "global",
                 "while", "class", "import", "try","except","from","break"]

py_other_kws = foxdot_kw + ["True", "False", "None", "self"]

py_key_types = ["abs","divmod","input","open","staticmethod","all","enumerate","int",
                "ord","str","any","eval","isinstance", "pow","sum","basestring",
                "execfile","issubclass","super","bin","file","iter","property","tuple",
                "bool", "filter","len","range","type","bytearray","float","list",
                "raw_input","unichr","callable","format","locals","reduce","unicode",
                "chr","frozenset","long","reload","vars","classmethod","getattr","map",
                "repr","xrange","cmp","globals","max","reversed","zip","compile","hasattr",
                "memoryview","round","__import__","complex","hash","min","delattr","set",
                "help","next","setattr","dict","hex","object","slice","dir","id","sorted"]

py_key_types = foxdot_funcs + py_key_types

py_separators = list("[](){},./*+=- \t\n")
py_whitespace = list(" \t\n\r\f\v")

left_b = list("([{'\"")
right_b = list(")]}'\"")
brackets = dict([(right_b[i],left_b[i]) for i in range(3)])
tabsize = 4

# RegEx Group 2

functions = r"(?<![a-zA-Z._])(" + "|".join(py_functions) + ")(?![_a-zA-Z])"
key_types = r"(?<![a-zA-Z._])(" + "|".join(py_key_types) + ")(?![_a-zA-Z])"
#key_types = r"\.?(\w+)\(|\W+P\W+" # TODO - change py_key_types to just and value followed by ()?
other_kws = r"(?<![a-zA-Z._])(" + "|".join(py_other_kws) + ")(?![_a-zA-Z])"

colour_map = {'plaintext'  : COLOURS.plaintext,
              'background' : COLOURS.background,
              'functions'  : COLOURS.functions,
              'key_types'  : COLOURS.key_types,
              'user_defn'  : COLOURS.user_defn,
              'other_kws'  : COLOURS.other_kws,
              'comments'   : COLOURS.comments,
              'numbers'    : COLOURS.numbers,
              'strings'    : COLOURS.strings,
              'dollar'     : COLOURS.dollar,
              'arrow'      : COLOURS.arrow,
              'players'    : COLOURS.players}

re_patterns = {  'functions' : functions,
                 'key_types' : key_types,
                 'user_defn' : user_defn,
                 'other_kws' : other_kws,
                 'numbers'   : numbers,
                 'strings'   : strings,
                 'dollar'    : dollar,
                 'arrow'     : arrow,
                 'players'   : players }


bracket_formatting = {'borderwidth': 2,
                      'relief' : 'groove'}

# Weightings (heavier get checked last)

tag_weights = [['numbers'],
               ['key_types','functions','user_defn','other_kws','dollar','arrow','players'],
               ['strings'],
               ['comments']]

def get_keywords():
    return list(set(py_indent_kw + py_functions + py_other_kws + py_key_types))



""" Allows you continually update a function """

from __future__ import absolute_import, division, print_function

# TODO: dependencies

_live_functions_dict = {}

class _live_function:
    def __init__(self, func, dependency=None):
        self.func = func
        self.name = func.__name__
        self.live = False
        self.dependency = dependency
    def __call__(self, *args, **kwargs):
        self.live = True
        if isinstance(self.dependency,  self.__class__):
            self.dependency.live = False
        return self.func.__call__(*args, **kwargs)
    def update(self, func, dependency=None):
        self.func = func
        if dependency:
            self.dependency = dependency
        return

def livefunction(f, dependency=None):
    """ Wraps a function 'f' in a flexible/interactive way """
    # Live functions can "depend" on others for switching from live or not
    if dependency in _live_functions_dict:
        dependenct = _live_functions_dict[dependency.__name__]
    # Add / update a dictionary of all live functions
    if f.__name__ not in _live_functions_dict:
        _live_functions_dict[f.__name__] = _live_function(f, dependency)
    else:
        _live_functions_dict[f.__name__].update(f, dependency)
    f = _live_functions_dict[f.__name__]
    # If the function is "live" call it
    if f.live: f.__call__()    
    return f


if __name__ == "__main__":
    # debug

    @livefunction
    def part1():
        return 10

    @livefunction
    def part2():
        return 20

    part1()

    print(part1.__class__)



""" Handles OSC messages being sent to SuperCollider.
"""
from __future__ import absolute_import, division, print_function
import sys

if sys.version_info[0] > 2:
    import queue
else:
    import Queue as queue

import json
import socket
import sys
import threading
import time
import itertools
import os.path

from collections import namedtuple
from threading import Thread

from .Code import WarningMsg
from .Settings import *

if sys.version_info[0] > 2:
    from .OSC3 import *
else:
    from .OSC import *

# Keep in sync with Info.scd
ServerInfo = namedtuple(
    'ServerInfo',
    ('sample_rate', 'actual_sample_rate', 'num_synths', 'num_groups',
     'num_audio_bus_channels', 'num_control_bus_channels',
     'num_input_bus_channels', 'num_output_bus_channels', 'num_buffers',
     'max_nodes', 'max_synth_defs'))


class OSCClientWrapper(OSCClient):
    error_printed=False
    def send(*args, **kwargs):
        """ Sends the message given but prints errors instead of raising them """
        try:
            OSCClient.send(*args, **kwargs)
        except OSCClientError as e:
            if not OSCClientWrapper.error_printed:
                print("Error sending message to SuperCollider server instance: make sure FoxDot quark is running and try again.")
                OSCClientWrapper.error_printed = True


class OSCConnect(OSCClientWrapper):
    """ An OSCClientWrapper that connects on initialisation """
    def __init__(self, address):
        OSCClientWrapper.__init__(self)
        self.connect(address)


class RequestTimeout(Exception):
    """ Raised if expecting a response from the server but received none """


class BidirectionalOSCServer(OSCServer):
    """
    This is a combination client/server

    The UDP server is necessary for receiving responses from the SCLang server
    when we query it with requests.

    Note that this is not thread-safe, as the receive() method can discard messages
    """
    def __init__(self, server_address=('localhost', 0), client=None, return_port=0):
        OSCServer.__init__(self, server_address, client, return_port)
        self._server_thread = None
        self.addDefaultHandlers()
        self.addMsgHandler('default', self._handle_message)
        self._response_queue = queue.Queue()
        self._printed_error = False

    def connect(self, addr):
        """ Connect to an address and start the server thread """
        self.client.connect(addr)
        self.start()

    def start(self):
        """ Start the server thread. """
        if self._server_thread is not None:
            return
        self._server_thread = threading.Thread(target=self.serve_forever)
        self._server_thread.setDaemon(True)
        self._server_thread.start()

    def stop(self):
        """ Stop the server thread and close the socket. """
        if self._server_thread is None:
            return
        self.running = False
        self._server_thread.join()
        self.server_close()

    def _handle_message(self, addr, tags, data, client_address):
        self._response_queue.put((addr, data))

    def send(self, *args, **kwargs):
        try:
            self.client.send(*args, **kwargs)
        except OSCClientError as e:
            if not self._printed_error:
                print("Error: No connection made to SuperCollider server instance.")
                self._printed_error = True

    def receive(self, pattern, timeout=2):
        """
        Retrieve the first message matching the pattern

        All messages received that do not match will be discarded
        """
        expr = getRegEx(pattern)
        now = start = time.time()
        while now - start < timeout:
            try:
                addr, data = self._response_queue.get(True, start + timeout - now)
            except queue.Empty:
                raise RequestTimeout()
            if type(addr) is bytes:
                addr = addr.decode()
            match = expr.match(addr)
            if match and (match.end() == len(addr)):
                return data
            now = time.time()

#  Create an abstract base class that could be sub-classed for users who want to send their OSC messages elsewhere

class ServerManager(object):
    def __init__(self, addr, port, osc_address="/s_new"):
        self.addr = addr
        self.port = port
        self.client = OSCClientWrapper()
        self.client.connect( (self.addr, self.port) )
        self.osc_address = osc_address

        self.node = 1000
        self.num_input_busses = 2
        self.num_output_busses = 2
        self.bus = self.num_input_busses + self.num_output_busses
        self.max_busses = 100
        self.max_buffers = 1024

    @staticmethod
    def create_osc_msg(dictionary):
        """ Converts a Python dictionary into an OSC style list """
        msg = []
        for key, value in dictionary.items():
            msg += [key, value]
        return msg

    def sendOSC(self, osc_message):
        self.client.send( osc_message )
        return

    def get_bundle(self, *args, **kwargs):
        bundle  = OSCBundle(time=kwargs.get("timestamp", 0))
        message = OSCMessage(self.osc_address)
        for item in args:
            if type(item) == dict:
                message.append(self.create_osc_msg(item))
            else:
                message.append(item)
        bundle.append(message)
        return bundle

    def loadSynthDef(self, *args, **kwargs):
        return
    def setFx(self, *args, **kwargs):
        return

class SCLangServerManager(ServerManager):

    fxlist    = None
    synthdefs = None

    def __init__(self, addr, osc_port, sclang_port):

        self.addr = addr
        self.port = osc_port
        self.SCLang_port = sclang_port

        self.midi_nudge = 0

        self.booted = False
        self.wait_time = 5
        self.count = 0

        # Assign a valid OSC Client
        self.forward = None

        self.node = 1000
        self.num_input_busses = 2
        self.num_output_busses = 2
        self.bus = self.num_input_busses + self.num_output_busses
        self.max_busses = 100
        self.max_buffers = 1024

        self.fx_setup_done = False
        self.fx_names = {}

        self.reset()

    def reset(self):

        # General SuperCollider OSC connection
        self.client = OSCClientWrapper()
        self.client.connect( (self.addr, self.port) )

        # OSC Connection for custom OSCFunc in SuperCollider
        if GET_SC_INFO:
            self.sclang = BidirectionalOSCServer()
            self.sclang.connect( (self.addr, self.SCLang_port) )
            self.loadSynthDef(FOXDOT_INFO_FILE)
            try:
                info = self.getInfo()
            except RequestTimeout:
                # It's not terrible if we couldn't fetch the info, but we should log it.
                WarningMsg("Could not fetch info from SCLang server. Using defaults...")
            else:
                self.max_buffers = info.num_buffers
                self.num_input_busses = info.num_input_bus_channels
                self.num_output_busses = info.num_output_bus_channels
                self.max_busses = info.num_audio_bus_channels
                self.bus = self.num_input_busses + self.num_output_busses
        else:
            self.sclang = OSCClientWrapper()
            self.sclang.connect( (self.addr, self.SCLang_port))

        # Clear SuperCollider nodes if any left over from other session etc

        self.freeAllNodes()

        # Load recorder OSCFunc

        self.loadRecorder() # move to the quark?

        # Toggle debug in SuperCollider

        self.dumpOSC(0)

    def __str__(self):
        return "FoxDot ServerManager Instance -> {}:{}".format(self.addr, self.port)

    def __repr__(self):
        return str(self)

    def nextnodeID(self):
        """ Gets the next node ID to use in SuperCollider """
        self.node += 1
        return self.node

    def query(self):
        """ Prints debug status to SuperCollider console """
        self.client.send(OSCMessage("/status"))
        return

    def nextbusID(self):
        """ Gets the next SuperCollider bus to use """
        self.bus += 2
        # Make sure we still have 2 audio channels available
        if self.bus + 1 >= self.max_busses:
            self.bus = self.num_input_busses + self.num_output_busses
        return self.bus

    def sendOSC(self, osc_message):
        """ Sends an OSC message to the server. Checks for midi messages """
        
        if osc_message.address == OSC_MIDI_ADDRESS:
        
            self.sclang.send( osc_message )
        
        else:
        
            self.client.send( osc_message )
        
        # If we are sending other messages as well
        
        if self.forward is not None:
            
            self.forward.send(osc_message)
        
        return

    def freeAllNodes(self):
        """ Triggers a free all message to kill all active nodes (sounds) in SuperCollider """
        msg = OSCMessage("/g_freeAll")
        msg.append([1])
        self.client.send(msg)
        return

    def setFx(self, fx_list):
        self.fxlist   = fx_list
        self.fx_names = {name: fx.synthdef for name, fx in fx_list.items() }
        return

    def set_midi_nudge(self, value):
        self.midi_nudge = value
        return

    def get_midi_message(self, synthdef, packet, timestamp):
        """ Prepares an OSC message to trigger midi sent from SuperCollider """

        bundle = OSCBundle(time=timestamp)
        bundle.setAddress(OSC_MIDI_ADDRESS) # these need to be variable names at least

        msg     = OSCMessage(OSC_MIDI_ADDRESS)

        note    = packet.get("midinote", 60)
        vel     = min(127, (packet.get("amp", 1) * 128) - 1)
        sus     = packet.get("sus", 0.5)
        channel = packet.get("channel", 0)
        nudge   = self.midi_nudge

        msg.append( [synthdef, note, vel, sus, channel, nudge] )

        bundle.append(msg)

        return bundle


    def get_init_node(self, node, bus, group_id, synthdef, packet):
    
        msg = OSCMessage("/s_new")

        # Make sure messages release themselves after 8 * the duration at max (temp)
        
        max_sus = float(packet["sus"] * 8) # might be able to get rid of this
        
        key = "rate" if synthdef.name in (SamplePlayer, LoopPlayer) else "freq"
        
        if key in packet:
        
            value = ["rate", packet[key]]
        
        else:
        
            value = []
        
        osc_packet = ["startSound", node, 0, group_id, 'bus', bus, "sus", max_sus] + value

        msg.append( osc_packet )
        
        return msg, node

    def get_control_effect_nodes(self, node, bus, group_id, packet):

        pkg = []

        # Go through effects and put together with child attributes

        for fx in self.fxlist.order[0]:

            if fx in packet and packet[fx] != 0:

                # this_effect = effects[fx] # old pre-prepared

                # prepare each effect here

                this_effect = self.prepare_effect(fx, packet)

                # Get next node ID
                node, last_node = self.nextnodeID(), node
            
                msg = OSCMessage("/s_new")
            
                osc_packet = [self.fx_names[fx], node, 1, group_id, 'bus', bus] + this_effect
            
                msg.append(osc_packet)
            
                pkg.append(msg)

        return pkg, node

    def get_synth_node(self, node, bus, group_id, synthdef, packet):
        
        msg = OSCMessage("/s_new")

        new_message = {}

        for key in packet:

            if key not in ("env", "degree"): # skip some attr

                try:

                    new_message[key] = float(packet[key]) # is this not already the case?

                except (TypeError, ValueError) as e:

                    WarningMsg( "Could not convert '{}' argument '{}' to float. Set to 0".format( key, packet[key] ))
                    new_message[key] = 0.0

        # Get next node ID

        node, last_node = self.nextnodeID(), node

        osc_packet = [synthdef.name, node, 1, group_id, synthdef.bus_name, bus] \
            + self.create_osc_msg(new_message)

        msg.append( osc_packet )

        return msg, node

    def get_pre_env_effect_nodes(self, node, bus, group_id, packet):

        pkg = []

        for fx in self.fxlist.order[1]:

            if fx in packet and packet[fx] != 0:

                this_effect = self.prepare_effect(fx, packet)

                # Get next node ID
                node, last_node = self.nextnodeID(), node
                msg = OSCMessage("/s_new")
                osc_packet = [self.fx_names[fx], node, 1, group_id, 'bus', bus] + this_effect
                msg.append( osc_packet )
                pkg.append(msg)
    
        return pkg, node

    def get_synth_envelope(self, node, bus, group_id, synthdef, packet):

        env_packet = {  "sus" : packet["sus"],
                        "amp" : packet["amp"] }

        for key in ("atk", "decay", "rel", "legato", "curve", "gain"):

            # Try and get from the player

            value = packet.get(key, None)

            # If it is absent or set to None, get default from Synth

            if value is None:

                value = synthdef.get_default_env(key)

            # Store

            env_packet[key] = value

        env = synthdef.get_default_env("env") if packet.get("env", None) is None else packet.get("env", None)

        try:

            dest = env.get_env_name()

        except AttributeError as e:

            # Set the curve value

            env_packet["curve"] = env
            dest = "BasicEnvelope"

        node, last_node = self.nextnodeID(), node
        msg = OSCMessage("/s_new")
        osc_packet = [dest, node, 1, group_id, 'bus', bus] + self.create_osc_msg(env_packet)
        msg.append( osc_packet )

        return msg, node

    def get_post_env_effect_nodes(self, node, bus, group_id, packet):

        pkg = []

        for fx in self.fxlist.order[2]:

            if fx in packet and packet[fx] != 0:

                this_effect = self.prepare_effect(fx, packet)

                # Get next node ID
                node, last_node = self.nextnodeID(), node
                msg = OSCMessage("/s_new")
                osc_packet = [self.fx_names[fx], node, 1, group_id, 'bus', bus] + this_effect
                msg.append( osc_packet )
                pkg.append(msg)

        return pkg, node

    def prepare_effect(self, name, packet):
        """ Finds the child attributes in packet and returns an OSC style list """
        data   = []
        effect = self.fxlist[name]
        for key in effect.args:
             data.append(key)
             data.append(float(packet.get(key, effect.defaults[key])))
        return data

    def get_exit_node(self, node, bus, group_id, packet):
        
        msg = OSCMessage("/s_new")
        node, last_node = self.nextnodeID(), node
        osc_packet = ['makeSound', node, 1, group_id, 'bus', bus, 'sus', float(packet["sus"])]
        msg.append( osc_packet )

        return msg, node

    def get_bundle(self, synthdef, packet, timestamp=0):
        """ Returns the OSC Bundle for a notew based on a Player's SynthDef, and event and effects dictionaries """ 

        # Create a specific message for midi

        if synthdef == "MidiOut": # this should be in a dict of synthdef to functions maybe? we need a "nudge to sync"

            return self.get_midi_message(synthdef, packet, timestamp)

        # Create a bundle
        
        bundle = OSCBundle(time=timestamp)

        # Get the actual synthdef object

        synthdef = self.synthdefs[synthdef]

        # Create a group for the note
        group_id = self.nextnodeID()
        msg = OSCMessage("/g_new")
        msg.append( [group_id, 1, 1] )
        
        bundle.append(msg)

        # Get the bus and SynthDef nodes
        this_bus  = self.nextbusID()
        this_node = self.nextnodeID()

        # synthdef.preprocess_osc(packet) # so far, just "balance" to multiply amp by 1

        # First node of the group (control rate)

        msg, this_node = self.get_init_node(this_node, this_bus, group_id, synthdef, packet)

        # Add effects to control rate e.g. vibrato        

        bundle.append( msg )

        pkg, this_node = self.get_control_effect_nodes(this_node, this_bus, group_id, packet)

        for msg in pkg:

            bundle.append(msg)

        # trigger synth

        msg, this_node = self.get_synth_node(this_node, this_bus, group_id, synthdef, packet)

        bundle.append(msg)

        # ORDER 1

        pkg, this_node = self.get_pre_env_effect_nodes(this_node, this_bus, group_id, packet)

        for msg in pkg:

            bundle.append(msg)

        # ENVELOPE

        # msg, this_node = self.get_synth_envelope(this_node, this_bus, group_id, synthdef, packet)

        # bundle.append( msg )

        # ORDER 2 (AUDIO EFFECTS)

        pkg, this_node = self.get_post_env_effect_nodes(this_node, this_bus, group_id, packet)
    
        for msg in pkg:
    
            bundle.append(msg)

        # OUT

        msg, _ = self.get_exit_node(this_node, this_bus, group_id, packet)

        bundle.append(msg)

        return bundle        

    def send(self, address, message):
        """ Sends message (a list) to SuperCollider """
        msg = OSCMessage(address)
        
        msg.append(message)
        
        self.client.send(msg)
        
        # If we are sending other messages as well
        
        if self.forward is not None:
        
            self.forward.send(message)
        
        return

    def free_node(self, node):
        """ Sends a message to SuperCollider to stop a specific node """
        message = OSCMessage("/n_free")
        message.append(node)
        self.client.send( message )
        return

    def bufferRead(self, path, bufnum):
        """ Sends a message to SuperCollider to read an audio file into a buffer """
        message = OSCMessage("/b_allocRead")
        message.append([bufnum, path])
        self.client.send( message )
        return

    def bufferFree(self, bufnum):
        """ Sends a message to SuperCollider to free a buffer """
        message = OSCMessage("/b_free")
        message.append([bufnum])
        self.client.send(message)

    def sendMidi(self, msg, cmd=OSC_MIDI_ADDRESS):
        """ Sends a message to the FoxDot class in SuperCollider to forward a MIDI message """
        msg.setAddress(cmd)
        self.sclang.send(msg)
        return

    def loadSynthDef(self, fn, cmd='/foxdot'):
        """ Sends a message to the FoxDot class in SuperCollider to load a SynthDef from file """
        msg = OSCMessage()
        msg.setAddress(cmd)
        msg.append(fn)
        self.sclang.send(msg)
        return

    def loadRecorder(self):
        """ Loads an OSCFunc that starts/stops recording to a set path """
        self.loadSynthDef(FOXDOT_RECORD_FILE)
        self._is_recording = False
        return

    def record(self, fn=None):
        """ Starts recording audio from SuperCollider """

        if self._is_recording is False:

            if fn is None:
                
                fn = "{}.aiff".format(get_timestamp())
            
            path = os.path.join(RECORDING_DIR, fn)

            msg = OSCMessage('/foxdot-record')
            msg.append([1, path])
            self.sclang.send(msg)

            self._is_recording = True
        
        return

    def stopRecording(self):
        """ Stops recording audio from SuperCollider """
        if self._is_recording is True:
            
            msg = OSCMessage('/foxdot-record')
            msg.append([0, ""]) # flag to stop recording
            self.sclang.send(msg)
            
            self._is_recording = False
        
        return

    def loadCompiled(self, fn):
        """ Sends a message to SuperCollider to load a compiled SynthDef file """
        msg = OSCMessage()
        msg.setAddress('/d_load')
        msg.append(fn)
        self.client.send(msg)

    def dumpOSC(self, value=1):
        """ Debug - Dumps OSC messages SCLang side """
        msg = OSCMessage("/dumpOSC")
        msg.append(value)
        self.client.send(msg)
        return

    def dumpTree(self, group_id=0, flag=0):
        """ Server will print the node tree """
        msg = OSCMessage("/g_dumpTree")
        msg.append([group_id, flag])
        self.client.send(msg)

    def getInfo(self):
        """ Fetch info about the SCLang server """
        msg = OSCMessage()
        msg.setAddress('/foxdot/info')
        self.sclang.send(msg)
        info = ServerInfo(*self.sclang.receive('/foxdot/info'))
        return info

    def start(self):
        """ Boots SuperCollider using `subprocess`"""

        if not self.booted:
            
            os.chdir(SC_DIRECTORY)
            
            print("Booting SuperCollider Server...")

            self.daemon = subprocess.Popen([SCLANG_EXEC, '-D', FOXDOT_STARTUP_FILE])

            os.chdir(USER_CWD)

            self.booted = True

        else:
            
            print("Warning: SuperCollider already running")
            
        return

    def makeStartupFile(self):
        ''' Boot SuperCollider and connect over OSC '''

        # 1. Compile startup file

        with open(FOXDOT_STARTUP_FILE, 'w') as startup:

            startup.write('''Routine.run {
            s.options.blockSize = 128;
            s.options.memSize = 131072;
            s.bootSync();\n''')

            files = [FOXDOT_OSC_FUNC, FOXDOT_BUFFERS_FILE]
            files = files + GET_SYNTHDEF_FILES() + GET_FX_FILES()
            
            for fn in files:

                f = open(fn)
                startup.write(f.read())
                startup.write("\n\n")

            startup.write("};")

        return

    def quit(self):
        if self.booted:
            self.client.send(OSCMessage("/quit"))
            sleep(1)
            self.daemon.terminate()
        if self._is_recording:
            self.stopRecording()
        return

    def add_forward(self, addr, port):
        self.forward = OSCClientWrapper()
        self.forward.connect( (addr, port) )

try:
    
    import socketserver

except ImportError:

    import SocketServer as socketserver


class Message:
    """ Wrapper for JSON messages sent to the server """
    def __init__(self, data):
        self.data = data
    def __str__(self):
        """ Prepares the json message to be sent with first 4 digits
            denoting the length of the message """
        packet = str(json.dumps(self.data, separators=(',',':')))
        length = "{:04d}".format( len(packet) )
        return length + packet
    def __len__(self):
        return len(str(self))
    def asString(self):
        return str(self)

def read_from_socket(sock):
    """ Reads data from the socket """
    # Get number single int that tells us how many digits to read
    try:
        bits = int(sock.recv(4).decode())
    except:
        return None
    if bits > 0:
        # Read the remaining data (JSON)
        data = sock.recv(bits).decode()
        # Convert back to Python data structure
        return json.loads(data)

def send_to_socket(sock, data):
    """ Converts Python data structure to JSON message and
        sends to a connected socket """
    msg = Message(data)
    # Get length and store as string
    msg_len, msg_str = len(msg), str(msg).encode()
    # Continually send until we know all of the data has been sent
    sent = 0
    while sent < msg_len:
        bits = sock.send(msg_str[sent:])
        sent += bits
    return

class ThreadedServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """ Base class """
    pass

class TempoServer(ThreadedServer):
    """ Used in TempoClock.py to connect to instances of FoxDot over a network. Sends
        bpm changes over the network. On initial request this sends the start_time value
        of the clock """

    def __init__(self, clock, port=57999):
        # tempo clock
        RequestHandler.metro  = self.metro = clock
        RequestHandler.master = self

        # Address information
        self.hostname = str(socket.gethostname())

        # Listen on any IP
        self.ip_addr  = "0.0.0.0"
        self.port     = int(port)

        # Public ip for server is the first IPv4 address we find, else just show the hostname
        self.ip_pub = self.hostname
        
        try:

            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            self.ip_pub = s.getsockname()[0]
            s.close()

        except OSError:

            pass

        # Instantiate server process

        self.peers = []

        ThreadedServer.__init__(self, (self.ip_addr, self.port), RequestHandler)
        self.server_thread = Thread(target=self.serve_forever)
        self.server_thread.daemon = False
        self.running = False

    def __str__(self):
        return "{} on port {}\n".format(self.ip_pub, self.port)

    def start(self):
        """ Starts listening on the socket """
        self.running = True
        self.server_thread.start()
        return

    def update_tempo(self, source, bpm, bpm_start_beat, bpm_start_time):
        """ Sends information  to all connected peers about changing tempo """
        for peer in self.peers:
            if peer is not source:
                peer.update_tempo(bpm, bpm_start_beat, bpm_start_time)
        # Update the master clock tempo if receiving from another peer
        if source is not None:
            self.metro.update_tempo_from_connection(bpm, bpm_start_beat, bpm_start_time)
        return

    def kill(self):
        """ Properly terminates the server instance """
        self.running = False
        self.server_thread.join(0)
        self.shutdown()
        self.server_close()
        return

class RequestHandler(socketserver.BaseRequestHandler):
    """ Created whenever a new connection to the server is made:
        self.request = socket
        self.server  = Server instance
        self.client_address = (address, port)
    """
    master = None

    def handle(self):
        """ Overload """

        print("New connection from {}".format(self.client_address))

        # First we get latency

        data = read_from_socket(self.request)

        # Should be "init" message

        assert "init" in data

        send_to_socket(self.request, {"clock_time": time.time()}) # maybe time at a beat?

        self.master.peers.append(self)

        while True:

            data = read_from_socket(self.request)

            # If a client disconnects, remove and print message

            if data is None:

                return self.disconnect()

            else:

                # Get the requested data and send to client

                if "request" in data:

                    send_to_socket(self.request, self.metro.get_sync_info())

                # Tell server to update tempo and update clients

                elif "new_bpm" in data:

                    self.master.update_tempo(self, **data["new_bpm"])

                elif "latency":

                    send_to_socket(self.request, ["latency"])

        return

    def disconnect(self):
        """ Prints a message to the master clock and removes a reference to this client """
        print("Client disconnected from {}".format(self.client_address))
        self.master.peers.remove(self)
        return 0

    def update_tempo(self, bpm, bpm_start_beat, bpm_start_time):

        data = {
            "new_bpm" :
                {
                    "bpm" : bpm,
                    "bpm_start_time" : bpm_start_time,
                    "bpm_start_beat" : bpm_start_beat
            }
        }

        send_to_socket(self.request, data)

        return


class TempoClient:
    def __init__(self, clock):
        self.metro = clock

        self.sync_keys = ("bpm_start_beat", "bpm_start_time", "bpm")

        self.server_hostname = None
        self.server_port     = None
        self.server_address  = None
        
        self.socket   = None

    def connect(self, hostname, port=57890):
        """ Connects to the server instance """

        # Get details of remote
        self.server_hostname = hostname
        self.server_port     = int(port)
        self.server_address  = (self.server_hostname, self.server_port)

        # Connect to remote

        try:

            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            self.socket.connect(self.server_address)

        except Exception as e:

            raise(e)

            raise(ConnectionError("Could not connect to host '{}'".format( self.server_hostname ) ) )

        # connect to the server and listen for new updates for the tempo-clock

        self.listening = True
        self.daemon = Thread(target=self.listen)
        self.daemon.start()

        # Send init message
        self.start_time = None
        self.stop_time  = None
        self.latency    = None
        self.recording_latency = False

        send_to_socket(self.socket, ["init"])
        
        self.start_timing()

        return self

    def start_timing(self):
        """ Starts an internal timer for calculating latency """
        self.start_time = time.time()

    def stop_timing(self):
        """ Stops the internal timer and calculates latency  """
        self.stop_time = time.time()
        self.calculate_latency(self.start_time, self.stop_time)

    def calculate_latency(self, start, end):
        """ Returns (and stores) the latency using the start and end time to send a message to the master server"""
        self.latency = (end - start) * 0.5
        return self.latency

    def record_latency(self):
        self.start_timing()
        self.recording_latency = True
        self.send(["latency"])
        return

    def send(self, data):
        """ Sends data to server """
        return send_to_socket(self.socket, data)

    def listen(self):
        """ Listens out for data coming from the server and passes it on
            to the handler.
        """
        
        # First message is machine clock time

        time_data = read_from_socket(self.socket)

        self.stop_timing()

        # self.metro.calculate_nudge(time_data["clock_time"], self.stop_time, self.latency)
        self.metro.calculate_nudge(time_data["clock_time"], self.start_time, self.latency)
        
        # Enter loop

        while self.listening:
            
            data = read_from_socket(self.socket)

            # Might be recording latency

            if self.recording_latency:

                self.stop_timing()
                self.recording_latency = False
            
            if data is None:
                break
            
            if "sync" in data:

                for key in self.sync_keys:
                    if key in data["sync"]:
                        object.__setattr__(self.metro, key, data["sync"][key])

                self.metro.update_tempo_from_connection(**data["sync"])

                self.metro.flag_wait_for_sync(False)
            
            elif "new_bpm" in data:

                self.metro.update_tempo_from_connection(**data["new_bpm"])
        return

    def update_tempo(self, bpm, bpm_start_beat, bpm_start_time):
        """ Sends data to other connected FoxDot instances to update their tempo """
        data = {
            "new_bpm" :
                {
                    "bpm" : bpm,
                    "bpm_start_time" : bpm_start_time,
                    "bpm_start_beat" : bpm_start_beat
            }
        }
        return self.send(data)

    def kill(self):
        """ Properly terminates the connection to the server """
        self.listening = False
        self.socket.close()
        return


if __name__ != "__main__":

    from .Settings import ADDRESS, PORT, PORT2, FORWARD_PORT, FORWARD_ADDRESS

    # DefaultServer = SCLangServerManager(ADDRESS, PORT, PORT2)
    Server = SCLangServerManager(ADDRESS, PORT, PORT2)

    if FORWARD_PORT and FORWARD_ADDRESS:
        Server.add_forward(FORWARD_ADDRESS, FORWARD_PORT)



"""    
    Players are what make FoxDot make music. They are similar in design to
    SuperCollider's `PDef` and `PBind` combo but with slicker syntax. FoxDot
    uses SuperCollider to *actually* make the sound and does so by triggering
    predefined `SynthDefs` - sort of like definitions of a digital instruments.
    To have a look at the list of `SynthDefs`, you can just `print` them to
    the console: ::

        print(SynthDefs)

    Each one of these represents a `SynthDef` *object*. These objects are then
    given to Players to play - like giving an instrument to someone in your
    orchestra. To give someone the instrument, `pads`, you use a double arrow
    some code syntax like this: ::

        p1 >> pads()

    To stop a Player, use the `stop` method e.g. `p1.stop()`. If you want to
    stop all players, you can use the command `Clock.clear()` or the keyboard
    shortcut `Ctrl+.`, which executes this command.

    `p1` is the name of a predefined player object. At startup, FoxDot reserves
    all one- and two-character variable names, such as `x`, `p1`, or `bd` for
    player objects but these can be repurposed if you like. If you want to use
    a variable name for a player object with more than two characters, you just
    instantiate a new `Player` object: ::

        foo = Player()

        foo >> pads()

    Changing parameters
    -------------------

    By default, player objects play the first note of their default scale (more
    below) with a duration of 1 beat per note. To change the pitch just give the
    `SynthDef` a list of numbers. ::

        p1 >> pads([0,7,6,4])

    Play multiple pitches together by putting them in round brackets: ::

        p1 >> pads([0,2,4,(0,2,4)])
    
    When you start FoxDot up, your clock is ticking at 120bpm and your player
    objects are all playing in the major scale. With 8 pitches in the major scale,
    the 0 refers to the first pitch and the 7 refers to the pitch one octave
    higher because Python, like most programming languages, uses zero-indexing.
    To change your scale you can specify a new scale as a keyword argument (see
    the documentation on `Scales` for more information on scales) or change the
    default scale for all player objects. ::

        # Changing scale as a keyword argument
        p1 >> pads([0,7,6,4], scale=Scale.minor)

        # Changing the default scalew (the following are equivalent)
        Scale.default.set("minor")
        Scale.default.set(Scale.minor)
        Scale.default.set([0,2,3,5,7,8,10])

        # See a list of scales
        print Scale.names()

        # Change the tempo (this takes effect at the next bar)
        Clock.bpm = 144

    To change the rhythm of your player object, specify the durations using
    the `dur` keyword. Other keywords can be specified, such as `oct` for the
    octave and `sus` for the sustain, which is the same as the duration by
    default. ::

        p1 >> pads([0,7,6,4], dur=[1,1/2,1/4,1/4], oct=6, sus=1)

        # See a list of possible keyword arguments
        print(Player.get_attributes())

    Using the `play` SynthDef
    -------------------------

    There is a special case SynthDef object called `play` which allows you
    to play short audio files rather than specify pitches. In this case
    you use a string of characters as the first argument where each character
    refers to a different folder of audio files. You can see more information
    by evaluating `print(Samples)`. The following line of code creates
    a basic drum beat: ::

        d1 >> play("x-o-")

    To play multiple patterns simultaneously, you can create a new `play` object. This
    is useful if you want to have different attributes for each player. ::
        
        bd >> play("x( x)  ", dur=1)
        hh >> play("---[--]", dur=[1/2,1/2,1/4], rate=4)
        sn >> play("  o ", rate=(.9,1), pan=(-1,1))

    Grouping characters in round brackets laces the pattern so that on each
    play through of the sequence of samples, the next character in the group's
    sample is played. The sequence `(xo)---` would be played back as if it
    were entered `x---o---`. Using square brackets will force the enclosed samples
    to played in the same time span as a single character e.g. `--[--]` will play
    two hi-hat hits at a half beat then two at a quarter beat. You can play a
    random sample from a selection by using curly braces in your Play String
    like so: ::

        d1 >> play("x-o{-[--]o[-o]}")

    FoxDot Player Object Keywords
    -----------------------------

    dur - Durations (defaults to 1 and 1/2 for the Sample Player)

    sus - Sustain (defaults to `dur`)

    amp - Amplitude (defaults to 1)

    rate - Variable keyword used for misc. changes to a signal. E.g. Playback rate of the Sample Player (defaults to 1)

    delay - A duration of time to wait before sending the information to SuperCollider (defaults to 0)

    sample - Special keyword for Sample Players; selects another audio file from the bank of samples for a sample character.
    

"""

from __future__ import absolute_import, division, print_function

import itertools
from functools import partial

from os.path import dirname
from random import shuffle, choice
from copy import copy, deepcopy

from .Settings import SamplePlayer, LoopPlayer
from .Code import WarningMsg, debug_stdout
from .SCLang.SynthDef import SynthDefProxy, SynthDef, SynthDefs
from .Effects import FxList
from .Utils import stdout
from .Buffers import Samples

from .Key import *
from .Repeat import *
from .Patterns import *
# from .Midi import *

from .Root import Root
from .Scale import Scale, ScaleType, ScalePattern
from .Scale import midi, miditofreq, get_freq_and_midi

from .Bang import Bang

from .TimeVar import TimeVar, Pvar

class EmptyPlayer(object):
    """ Place holder for Player objects created at run-time to reduce load time.
    """
    def __init__(self, name):
        self.name = name
    def __repr__(self):
        return "<{} - Unassigned>".format(self.name)
    
    def __rshift__(self, *args, **kwargs):
        """ Converts an EmptyPlayer to a Player. """
        self.__class__ = Player
        self.__init__(self.name)
        self.__rshift__(*args, **kwargs)
        return self

    def __invert__(self):
        return self.reset()

    def __getattribute__(self, name):
        """ Tries to return the correct attr; if not init the Player and try again """
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            self.__class__ = Player
            self.__init__(self.name)
            try:
                return self.__getattribute__(name)
            except AttributeError:
                return self.__getattr__(name) # use getattr to make sure we return player key


class Player(Repeatable):

    """
    FoxDot generates music by creating instances of `Player` and giving them instructions
    to follow. At startup FoxDot creates many instances of `Player` and assigns them to
    any valid two character variable. This is so that when you start playing you don't 
    have to worry about typing `myPlayer = Player()` and `myPlayer_2 = Player()` every
    time you want to do something new. Of course there is nothing stopping you from 
    doing that if yo so wish.

    Instances of `Player` are given instructions to generate music using the `>>` syntax,
    overriding the bitshift operator, and should be given an instance of `SynthDefProxy`.
    A `SynthDefProxy` is created when calling an instance of `SynthDef` - these are the
    "instruments" used by player objects and are written in SuperCollider code. You can
    see more information about these in the `SCLang` module. Below describes how to assign
    a `SynthDefProxy` of the `SynthDef` `pads` to a `Player` instance called `p1`: ::

        # Calling pads as if it were a function returns a 
        # pads SynthDefProxy object which is assigned to p1
        p1 >> pads()

        # You could store several instances and assign them at different times
        proxy_1 = pads([0,1,2,3], dur=1/2)
        proxy_2 = pads([4,5,6,7], dur=1)

        p1 >> proxy_1 # Assign the first to p1
        p1 >> proxy_2 # This replaces the instructions being followed by p1
    """

    # Set private values

    debug = 0

    __vars = []
    __init = False

    # Really need to tidy this up

    keywords   = ('degree', 'oct', 'freq', 'dur', 'delay', 'buf',
                  'blur', 'amplify', 'scale', 'bpm', 'sample', "env")

    envelope_keywords = ("atk", "decay", "rel", "legato", "curve", "gain")

    # Base attributes
    
    base_attributes = ('sus', 'fmod', 'pan', 'rate', 'amp', 'midinote', 'channel') 

    required_keys = ("amp", "sus")
    
    internal_keywords = tuple(value for value in keywords if value != "degree")

    # Aliases

    alias = { "pitch" : "degree",
              "char"  : "degree" }

    fx_attributes = FxList.all_kwargs()
    fx_keys       = FxList.kwargs()

    # Load default sample bank
    samples = Samples

    # Set in __init__.py
    metro   = None

    default_scale = Scale.default

    default_root  = Root.default() # TODO//remove callable

    after_update_methods = ["stutter"]

    # Tkinter Window
    widget = None

    def __init__(self, name=None):

        # Inherit from repeatable i.e. x.every

        Repeatable.__init__(self)

        self.method_synonyms["->"] = "rshift"
        self.method_synonyms["<-"] = "lshift"
    
        # General setup
        
        self.synthdef = None
        self.id = name

        #self.current_event_size   = 0
        #self.current_event_length = 0
        #self.current_event_depth  = 0

        # Stopping flag
        self.stopping = False
        self.stop_point = 0

        # Reference to other objects in the clock played at the same time
        self.queue_block = None
        self.bus = None

        # The string representation of the degree of the player
        self.playstring = ""

        # Information used in generating OSC messages
        self.buf_delay = []
        self.timestamp = 0
        # self.condition = lambda: True
        # self.sent_messages = []

        # Visual feedback information

        self.envelope    = None
        self.line_number = None
        self.whitespace  = None
        self.do_bang     = False
        self.bang_kwargs = {}

        # Keeps track of which note to play etc

        self.event_index = 0
        self.event_n = 0
        self.notes_played = 0
        self.event = {}
        self.accessed_keys = []

        # Used for checking clock updates

        self.current_dur = None
        self.old_pattern_dur = None
        self.old_dur = None
        
        self.isplaying = False
        self.isAlive = True

        # These dicts contain the attribute and modifier values that are sent to SuperCollider     

        self.attr  = {}
        self.modifier = Pattern()
        self.mod_data = 0
        self.filename = None

        # Keyword arguments that are used internally

        self.scale = None
        self.offset  = 0
        self.following = None
        
        # List the internal variables we don't want to send to SuperCollider

        self.__vars = list(self.__dict__.keys())
        self.__init = True

        self.reset()

    # Class methods

    @classmethod
    def help(cls):
        return print(cls.__doc__)

    @classmethod
    def get_attributes(cls):
        """ Returns a list of possible keyword arguments for FoxDot players and effects """
        return cls.keywords + cls.base_attributes + cls.fx_attributes

    @classmethod
    def Attributes(cls):
        """ To be replaced by `Player.get_attributes()` """
        return cls.get_attributes()

    @classmethod
    def set_clock(cls, tempo_clock):
        cls.metro = tempo_clock

    # Should this also be instance method?
    @classmethod
    def set_sample_bank(cls, sample_bank):
        cls.samples = sample_bank

    def __hash__(self):
        return hash(self.id) # could be problematic if there are id clashes?

    # Player Object Manipulation
    
    def __rshift__(self, other):
        """ Handles the allocation of SynthDef objects using >> syntax, other must be
            an instance of `SynthDefProxy`, which is usually created when calling a
            `SynthDef`
        """
        
        if isinstance(other, SynthDefProxy):
            
            # Call the update method
            
            self.update(other.name, other.degree, **other.kwargs)
        
            # self.update_pattern_root('sample' if self.synthdef == SamplePlayer else 'degree')
            
            for method, arguments in other.methods:
            
                args, kwargs = arguments
            
                getattr(self, method).__call__(*args, **kwargs)
            
            # Add the modifier (check if not 0 to stop adding 0 to values)

            if (not isinstance(other.mod, (int, float))) or (other.mod != 0):

                self + other.mod
            
            return self
        
        raise TypeError("{} is an innapropriate argument type for PlayerObject".format(other))

        return self

    def test_for_circular_reference(self, value, attr, last_player=None, last_attr=None):
        """ Used to raise an exception if a player's attribute refers to itself e.g. `p1 >> pads(dur=p1.dur)` """

        # We are setting self.attr to value, check if value depends on self.attr

        if isinstance(value, PGroup):

            for item in value:

                self.test_for_circular_reference(item, attr, last_player, last_attr)

        elif isinstance(value, PlayerKey):

            # If the Player key relies on this player.attr, raise error

            if value.cmp(self, attr):

                ident_self = value.name()

                if last_player is not None:

                    ident_other = "{}.{}".format(last_player.id, last_attr)

                else:

                    ident_other = ident_self

                err = "Circular reference found: {} to itself via {}".format(ident_self, ident_other)

                raise ValueError(err)

            elif last_player == value.player and last_attr == value.attr:

                return

            else:

                # Go through the player key's 

                for item in value.get_player_attribute():
            
                    self.test_for_circular_reference(item, attr, value.player, value.attr)

        return

    def __setattr__(self, name, value):

        # Possibly replace with slots?

        if self.__init:

            # Force the data into a Pattern if the attribute is used with SuperCollider
            
            if name not in self.__vars:

                # Get any alias

                name = self.alias.get(name, name)

                value = asStream(value)

                for item in value:

                    self.test_for_circular_reference(item, name)

                # Update the attribute dict if no error
                
                self.attr[name] = value

                # Remove from the stored pattern dict / call those

                self.update_pattern_root(name)

                # keep track of what values we change with +-

                if (self.synthdef == SamplePlayer and name == "sample") or (self.synthdef != SamplePlayer and name == "degree"):

                    self.modifier = value

                # Update any playerkey

                if name in self.__dict__:

                    if isinstance(self.__dict__[name], PlayerKey):

                        self.__dict__[name].update_pattern()

                # self.update_player_key(name, 0, 0)

                return
            
        self.__dict__[name] = value

        return

    def __getattr__(self, name):
        try:
            # This checks for aliases, not the actual keys
            name = self.alias.get(name, name)

            if name in self.attr and name not in self.__dict__:

                # Return a Player key

                self.update_player_key(name, self.now(name), 0)

            item = self.__dict__[name]

            # If returning a player key, keep track of which are being accessed
        
            if isinstance(item, PlayerKey) and name not in self.accessed_keys:                
        
                self.accessed_keys.append(name)
            
            return item
        
        except KeyError:
            
            err = "Player Object has no attribute '{}'".format(name)
            
            raise AttributeError(err)

        return

    def __getattribute__(self, name):
        # This checks for aliases, not the actual keys
        name = Player.alias.get(name, name)
        item = object.__getattribute__(self, name)
        if isinstance(item, PlayerKey):
            if name not in self.accessed_keys:
                self.accessed_keys.append(name)
        return item

    def __getitem__(self, name):
        if self.__init:
            if name not in self.__vars:
                return self.attr[name]
            pass
        return self.__dict__[name]

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return not self is other

    # --- Startup methods

    def reset(self):
        """ Sets all Player attributes to 0 unless their default is specified by an effect. Also
            can be called by using a tilde before the player variable. E.g. ~p1 """

        # Add all keywords to the dict, then set non-zero defaults

        reset = []

        for key in Player.Attributes():

            if key not in ("scale", "dur", "sus", "blur", "amp",
                            "amplify", "degree", "oct", "bpm"):

                setattr(self, key, 0)

            reset.append(key)

        # Set any non zero defaults for effects, e.g. verb=0.25

        for key in Player.fx_attributes:

            value = FxList.defaults[key]

            setattr(self, key, value)

            reset.append(key)

        # Set SynthDef defaults

        if self.synthdef in SynthDefs:

            synth = SynthDefs[self.synthdef]
            
            for key in ("atk", "decay", "rel"):

                setattr(self, key, synth.defaults[key])

                reset.append(key)

        # Any other attribute that might have been used - set to 0

        for key in self.attr:

            if key not in reset:

                setattr(self, key, 0)

        # Set any non-zero values for FoxDot

        # Sustain & Legato
        self.sus     = 0.5 if self.synthdef == SamplePlayer else 1
        self.blur    = 1

        # Amplitude
        self.amp     = 1
        self.amplify = 1

        # Duration of notes
        self.dur     = 0.5 if self.synthdef == SamplePlayer else 1

        # Degree of scale / Characters of samples
        self.degree  = " " if self.synthdef is SamplePlayer else 0

        # Octave of the note
        self.oct     = 5
        
        # Tempo
        self.bpm     = None

        # Stop calling any repeating methods

        self.stop_calling_all()
        
        return self

    def __invert__(self):
        """ Using the ~ syntax resets the player """
        return self.reset()

    # --- Update methods

    def __call__(self, **kwargs):
        """ Sends the next osc message event to SuperCollider and schedules this
            Player in the clock based on the current clock time and this player's
            current duration value. """

        # If stopping, kill the event

        if self.stopping and self.metro.now() >= self.stop_point:
            
            self.kill()
            
            return

        # If the duration has changed, work out where the internal markers should be

        # -- This could be in its own private function

        force_count = kwargs.get("count", False)
        dur_updated = self.dur_updated() 

        if dur_updated or force_count is True:

            try:

                self.event_n, self.event_index = self.count(self.event_index if not force_count else None)

            except TypeError as e:

                print(e)

                print("TypeError: Innappropriate argument type for 'dur'")

        # Get the current state 

        self.get_event() 

        # Play the note
        
        if not isinstance(self.event["dur"], rest):

            try:
        
                self.send(verbose=(self.metro.solo == self and kwargs.get('verbose', True)))

            except Exception as err:

                print("Error in Player {}: {}".format(self.id, err))
        
        # If using custom bpm

        dur = self.event["dur"]

        if self.event['bpm'] is not None:

            try:

                tempo_shift = float(self.metro.bpm) / float(self.event['bpm'])

            except (AttributeError, TypeError, ZeroDivisionError):

                tempo_shift = 1

            dur *= tempo_shift

        # Schedule the next event (could move before get_event and use the index for get_event)

        self.event_index = self.event_index + dur

        self.metro.schedule(self, self.event_index, kwargs={})

        # Change internal marker

        self.event_n += 1 
        self.notes_played += 1

        return

    def count(self, time=None, event_after=False):
        """ Counts the number of events that will have taken place between 0 and `time`. If
            `time` is not specified the function uses self.metro.now(). Setting `event_after`
            to `True` will find the next event *after* `time`"""

        n = 0
        acc = 0
        dur = 0
        now = (time if time is not None else self.metro.now())

        if self.current_dur is None:

            self.current_dur = self.rhythm()

        durations = list(map(get_first_item, self.current_dur)) # careful here
        
        total_dur = float(sum(durations))

        if total_dur == 0:

            WarningMsg("Player object has a total duration of 0. Set to 1")

            durations = [1]
            total_dur =  1 
            self.dur  =  1
    
        acc = now - (now % total_dur)

        try:

            n = int(len(durations) * (acc / total_dur))

        except TypeError as e:

            WarningMsg(e)

            self.stop()

            return 0, 0

        if acc != now:

            while True:

                dur = float(modi(durations, n))

                if acc + dur == now:

                    acc += dur

                    n += 1

                    break

                elif acc + dur > now:

                    if event_after:

                        acc += dur
                        n += 1

                    break

                else:
                    
                    acc += dur
                    n += 1

        # Returns value for self.event_n and self.event_index

        return n, acc

    def dur_updated(self):
        """ Returns True if the players duration has changed since the last call """
        self.current_dur = self.rhythm()
        if self.current_dur != self.old_dur:
            self.old_dur = self.current_dur
            return True
        return False

    def rhythm(self):
        """ Returns the players array of durations at this point in time """
        return list(map(lambda x: x if isinstance(x, (int, float)) else self.unpack(x), self.attr["dur"]))

    def update(self, synthdef, degree, **kwargs):
        """ Updates the attributes of the player. Called using the >> syntax.
        """

        # SynthDef name
        
        self.synthdef = synthdef

        # Make sure all values are reset to start

        if "filename" in kwargs:

            self.filename = kwargs["filename"]
            del kwargs["filename"]

        if self.isplaying is False:

            self.reset() 

        # If there is a designated solo player when updating, add this at next bar
        
        if self.metro.solo.active() and self.metro.solo != self:

            self.metro.schedule(lambda *args, **kwargs: self.metro.solo.add(self), self.metro.next_bar())

        # Update the attribute values

        special_cases = ["scale", "root", "dur"]

        # Set the degree

        if synthdef == SamplePlayer:

            if type(degree) == str:

                self.playstring = degree

            else:

                self.playstring = None

            if degree is not None:

                setattr(self, "degree", degree if degree != "" else " ")

        elif degree is not None:

            self.playstring = str(degree) # this doesn't work for var!

            setattr(self, "degree", degree)

        else:

            setattr(self, "degree", 0)            


        # Set special case attributes

        self.scale = kwargs.get("scale", self.__class__.default_scale )
        self.root  = kwargs.get("root",  self.__class__.default_root )

        # If only duration is specified, set sustain to that value also

        if "dur" in kwargs:

            # If we use tuples / PGroups in setting duration, use it to modify delay using the PDur algorithm

            setattr(self, "dur", kwargs["dur"])

            if "sus" not in kwargs:

                self.sus = self.attr['dur']

        # Set any other attributes

        for name, value in kwargs.items():

            if name not in special_cases:

                setattr(self, name, value)

        # Calculate new position if not already playing

        if self.isplaying is False:

            # Add to clock
            
            self.isplaying = True
            self.stopping = False

            # If we want to update now, set the start point to now

            after = True

            if self.metro.now_flag:

                start_point = self.metro.now()

                after = False

            elif kwargs.get("quantise", True) == False:

                start_point = self.metro.now()

            else:
            
                start_point = self.metro.next_bar()

            self.event_n = 0

            self.event_n, self.event_index = self.count(start_point, event_after=after)

            self.metro.schedule(self, self.event_index)

        return self

    def often(self, *args, **kwargs):
        """ Calls a method every 1/2 to 4 beats using `every` """
        return self.every(PRand(1, 8)/2, *args, **kwargs)

    def sometimes(self, *args, **kwargs):
        """ Calls a method every 4 to 16 beats using `every` """
        return self.every(PRand(8, 32)/2, *args, **kwargs)

    def rarely(self, *args, **kwargs):
        """ Calls a method every 16 to 32 beats using `every` """
        return self.every(PRand(32, 64)/2, *args, **kwargs)

    def get_timestamp(self, beat=None):
        if beat is not None:
            timestamp = self.metro.osc_message_time() - self.metro.beat_dur(self.metro.now() - beat)
        else:
            timestamp = self.metro.osc_message_time()
        return timestamp

    def stutter(self, amount=None, _beat_=None, **kwargs):
        """ Plays the current note n-1 times. You can specify keywords. """

        timestamp = self.get_timestamp(_beat_)
        
        # Get the current values (this might be called between events)

        n = int(kwargs.get("n", amount if amount is not None else 2))
        ahead = int(kwargs.get("ahead", 0))

        for key in ("ahead", "n"):
            if key in kwargs:
                del kwargs[key]

        # Only send if n > 1 and the player is playing

        if self.metro.solo == self and n > 1:

            new_event = {}

            attributes = self.attr.copy()
        
            attr_keys = set(list(self.attr.keys()) + list(kwargs.keys()))
            
            for key in attr_keys:

                if key in kwargs:

                    item = kwargs[key]

                    if isinstance(item, PGroupPrime):

                        new_event[key] = self.unpack(item)

                    elif isinstance(item, PGroup):

                        new_event[key] = self.unpack(PGroup([item]))

                    else:

                        new_event[key] = self.unpack(PGroup(item))

                elif len(attributes[key]) > 0:

                    new_event[key] = self.now(key, ahead)

            new_event = self.unduplicate_durs(new_event)

            dur = float(kwargs.get("dur", new_event["dur"])) / n

            new_event["dur"] = dur

            # Get PGroup delays

            new_event["delay"] = PGroup([dur * (i+1) for i in range(max(n, self.get_event_length(new_event)))])

            new_event = self.get_prime_funcs(new_event)

            self.send(timestamp=timestamp, **new_event)
                
        return self

    def jump(self, ahead=1, _beat_=None, **kwargs):
        """ Plays an event ahead of time. """

        timestamp = self.get_timestamp(_beat_)

        if self.metro.solo == self:

            new_event = {}
        
            attributes = copy(self.attr)
            
            for key in attributes:

                if key in kwargs:

                    new_event[key] = self.unpack(PGroup(kwargs[key]))

                elif len(attributes[key]) > 0:

                    new_event[key] = self.now(key, ahead)

            new_event = self.unduplicate_durs(new_event)

            new_event = self.get_prime_funcs(new_event)

            self.send(timestamp=timestamp, **new_event)
                
        return self

    def lshift(self, n=1):
        """ Plays the event behind """
        self.event_n -= (n+1)
        return self

    def rshift(self, n=1):
        """ Plays the event in front """
        self.event_n += n
        return self

    # Preset methods

    def spread(self, on=0.125):
        """ Sets pan to (-1, 1) and pshift to (0, 0.125)"""
        if on != 0:
            self.pan=(-1,1)
            self.pshift=(0,on)
        else:
            self.pan=0
            self.pshift=0
        return self

    def unison(self, unison=2, detune=0.125):
        """ Like spread(), but can specify number of voices(unison)  
        Sets pan to (-1,-0.5,..,0.5,1) and pshift to (-0.125,-0.0625,...,0.0625,0.125)
        If unison is odd, an unchanged voice is added in the center
        Eg : p1.unison(4, 0.5) => pshift=(-0.5,-0.25,0.25,0.5), pan=(-1.0,-0.5,0.5,1.0)
             p1.unison(5, 0.8) => pshift=(-0.8,-0.4,0,0.4,0.8), pan=(-1.0,-0.5,0,0.5,1.0)
        """
        if unison != 0:
            pan=[]
            pshift=[]
            uni = int(unison if unison%2==0 else unison-1)
            for i in range(1,int(uni/2)+1):
                pan.append(2*i/uni)
                pan.insert(0, -2*i/uni)
            for i in range(1, int(uni/2)+1):
                pshift.append(detune*(i/(uni/2)))
                pshift.insert(0,detune*-(i/(uni/2)))
            if unison%2!=0 and unison > 1:
                pan.insert(int(len(pan)/2), 0)
                pshift.insert(int(len(pan)/2), 0)              
            self.pan = tuple(pan)
            self.pshift = tuple(pshift)
        else:
            self.pan=0
            self.pshift=0
        return self

    def seconds(self):
        """ Sets the player bpm to 60 so duration will be measured in seconds """
        self.bpm=60
        return self
    
    def slider(self, start=0, on=1):
        """ Creates a glissando effect between notes """
        if on:
            if start:
                self.slide=[1,0]
                self.slidefrom=[0,1]
            else:
                self.slide=[0,1]
                self.slidefrom=[1,0]
            self.slidedelay=0.75
        else:
            self.slide=0
            self.slidefrom=0
            self.slidedelay=0
        return self

    def penta(self, switch=1):
        """ Shorthand for setting the scale to the pentatonic mode of the default scale """
        if switch:
            self.scale = self.__class__.default_scale.pentatonic
        else:
            self.scale = self.__class__.default_scale
        return self

    def alt_dur(self, dur):
        """ Used to set a duration that changes linearly over time. You should use a `linvar` but
            any value can be used. This sets the `dur` to 1 and uses the `bpm` attribute to 
            seemingly alter the durations """

        self.dur = 1
        self.bpm = self.metro.bpm*(1/(dur))
        return self

    def reverse(self):
        """ Reverses every attribute stream """
        for attr in self.attr:
            try:
                self.attr[attr] = self.attr[attr].pivot(self.event_n)
            except AttributeError:
                pass
        return self

    def shuffle(self):
        """ Shuffles the degree of a player. """
        # If using a play string for the degree
        #if self.synthdef == SamplePlayer and self.playstring is not None:
        #    # Shuffle the contents of playgroups among the whole string
        #    new_play_string = PlayString(self.playstring).shuffle()
        #    new_degree = Pattern(new_play_string).shuffle()
        #else:            
        #new_degree = self.attr['degree'].shuffle()
        new_degree = self.previous_patterns["degree"].root.shuffle()
        self._replace_degree(new_degree)
        return self


    def rotate(self, n=1):
        """ Rotates the values in the degree by 'n' """
        #self._replace_degree(self.attr['degree'].rotate(n))
        new_degree = self.previous_patterns["degree"].root.rotate(n)
        self._replace_degree(new_degree)
        return self

    def attrmap(self, key1, key2, mapping):
        """ Sets the attribute for self.key2 to self.key1
            altered with a mapping dictionary.
        """
        self.attr[key2] = self.attr[key1].map(mapping)
        return self

    def smap(self, kwargs):
        """ Like map but maps the degree to the sample attribute
        """
        self.attrmap("degree", "sample", kwargs)
        return self

    def map(self, other, mapping, otherattr="degree"):
        """ p1 >> pads().map(b1, {0: {oct=[4,5], dur=PDur(3,8), 2: oct})     """
        # Convert dict to {"oct": {4}}
        # key is the value of player key, attr is 
        for key, minimap in mapping.items():
            for attr, value in minimap.items():
                setattr(self, attr, mapvar(getattr(other, attr), values))
        return self

    
    # --- Misc. Standard Object methods

    def __int__(self):
        return int(self.now('degree'))

    def __float__(self):
        return float(self.now('degree'))

    def __add__(self, data):
        """ Change the degree modifier stream """
        self.mod_data = data
        if self.synthdef == SamplePlayer:
            # self.attr['sample'] = self.modifier + self.mod_data
            self.sample = self.modifier + self.mod_data
        else:
            #self.attr['degree'] = self.modifier + self.mod_data
            self.degree = self.modifier + self.mod_data
        return self

    def __sub__(self, data):
        """ Change the degree modifier stream """
        self.mod_data = 0 - data
        if self.synthdef == SamplePlayer:
            self.attr['sample'] = self.modifier + self.mod_data
        else:
            self.attr['degree'] = self.modifier + self.mod_data
        return self

    def __mul__(self, data):
        return self

    def __div__(self, data):
        return self

    # --- Data methods

    def __iter__(self):
        for _, value in self.event.items():
            yield value

    def number_of_layers(self, **kwargs):
        """ Returns the deepest nested item in the event """
        num = 1
        for attr, value in self.event.items():
            value = kwargs.get(attr, value)
            if isinstance(value, PGroup):
                l = pattern_depth(value)
            else:
                l = 1                
            if l >  num:
                num = l
        return num                

    def largest_attribute(self, **kwargs):
        """ Returns the length of the largest nested tuple in the current event dict """

        size = 1
        values = []

        for attr, value in self.event.items():
            value = kwargs.get(attr, value)
            l = get_expanded_len(value)
            if l > size:
                size = l
        return size

    def get_event_length(self, event=None, **kwargs):
        """ Returns the largest length value in the event dictionary """
        if event is None:
            
            event = self.event

        if kwargs:

            event = event.copy()
            event.update(kwargs)
        
        max_val = 0

        for attr, value in event.items():
            
            if isinstance(value, PGroup):
            
                l = len(value)
            
            else:
            
                l = 1

            if l > max_val:
                
                max_val = l

        return max_val

    def number_attr(self, attr):
        """ Returns true if the attribute should be a number """
        return not (self.synthdef == SamplePlayer and attr in ("degree", "freq"))

    def update_player_key(self, key, value, time):
        """  Forces object's dict uses PlayerKey instances
        """
        if (key not in self.__dict__) or (not isinstance(self.__dict__[key], PlayerKey)):

            self.__dict__[key] = PlayerKey(value, player=self, attr=key) 

        else:

            # Force values if not playing

            if self.isplaying is False:

                self.__dict__[key].set(value, time)

            else:

                self.__dict__[key].update(value, time)

        return

    def update_all_player_keys(self, ignore=[], event=None, **kwargs):
        """ Updates the internal values of player keys that have been accessed e.g. p1.pitch. If there is a delay,
            then schedule a function to update the values in the future. """

        # Don't bother if no keys are being accessed

        if len(self.accessed_keys) == 0:

            return

        if event is None:

            event = self.event

        delay = event.get("delay", 0)

        if isinstance(delay, PGroup):

            event_size = self.get_event_length(event, **kwargs)

            delays = itertools.cycle(delay)

            for i in range(event_size):

                delay = next(delays)

                # recursively unpack
                    
                new_event = {}

                for new_key, new_value in event.items():

                    if new_key in self.accessed_keys:

                        new_value = kwargs.get(new_key, new_value)

                        if isinstance(new_value, PGroup):

                            new_event[new_key]  = new_value[i]

                        else:

                            new_event[new_key] = new_value

                if isinstance(delay, PGroup):

                    # Recursively unpack and send messages

                    for i in range(self.get_event_length(new_event)):

                        self.update_all_player_keys(event=new_event, ignore=ignore, **kwargs)

                else:

                    self.update_player_key_from_event(new_event, time=self.event_index, ignore=ignore, delay=delay, **kwargs)

        else:

            self.update_player_key_from_event(event, time=self.event_index, ignore=ignore, delay=delay, **kwargs)

        return

    def update_player_key_from_event(self, event, time=None, delay=0, ignore=[], **kwargs):

        timestamp = self.event_index if time is None else time

        if delay == 0:

            for key in (x for x in self.accessed_keys if x not in ignore):

                self.update_player_key(key, kwargs.get(key, event.get(key, 0)), timestamp)

        else:

            func_args = (event, timestamp + delay, 0, ignore)

            self.metro.schedule(self.update_player_key_from_event, timestamp + delay, args=func_args, kwargs=kwargs)

        return

    def update_player_key_relation(self, item):
        """ Called during 'now' to update any Players that a player key is related to before using that value """

        # If this *is* the parent, just get the current value

        if item.parent is self:

            self.update_player_key(item.attr, self.now(item.attr), 0)

        # If the parent is in the same queue block, make sure its values are up-to-date

        elif self.queue_block is not None:

            # Try and find the item in the queue block

            try:

                queue_item = self.queue_block[item.player]

            except KeyError:

                queue_item = None

            # Update the parent with an up-to-date value

            if queue_item is not None and queue_item.called is False:
                    
                item.player.update_player_key(item.attr, item.player.now(item.attr), 0)

        return item.now()

    # --- Methods for preparing and sending OSC messages to SuperCollider

    def unpack(self, item):
        """ Converts a pgroup to floating point values and updates and time var or playerkey relations """

        if isinstance(item, GeneratorPattern):

            # "pop" value from the generator

            item = item.getitem() # could be renamed to "next"

        if isinstance(item, TimeVar):

            # Get current value if TimeVar

            item = item.now()

        if isinstance(item, NumberKey):

            # Update any relationships to the number key if necessary

            item = self.update_player_key_relation(item)

        if isinstance(item, PGroup):

            # Make sure any values in the PGroup have their "now" methods called

            item = item.convert_data(self.unpack)

        return item

    def get_key(self, key, i, **kwargs):
        return group_modi(kwargs.get(key, self.event[key]), i)

    # Private method

    def now(self, attr="degree", x=0, **kwargs):
        """ Calculates the values for each attr to send to the server at the current clock time """

        index = self.event_n + x

        try:

            if len(self.attr[attr]) > 0:

                attr_value = kwargs.get(attr, self.attr[attr][index])

            else:

                attr_value = 0 # maybe have a dict of defaults?

        # Debugging

        except KeyError as e:

            print(attr, self.attr[attr], index)
            raise(e)

        except ZeroDivisionError as e:

            print(self, attr, self.attr[attr], index)
            raise(e)

        # Force and timevar etc into floats

        if attr_value is not None and (not isinstance(attr_value, (int, float))):

            attr_value = self.unpack(attr_value)

        return attr_value

    def get_prime_funcs(self, event):
        """ Finds and PGroupPrimes in event and returns the modulated event dictionary """

        prime_keys = ("degree", "sample")

        # Go through priority keys

        for key in prime_keys:

            self.apply_prime_funcs(event, key)

        # Then do the rest (skipping prime)

        for key in event:

            if key not in prime_keys:

                self.apply_prime_funcs(event, key)

        return event

    @staticmethod
    def apply_prime_funcs(event, key):
        value = event[key]
        if isinstance(value, PGroup) and value.has_behaviour():
            func = value.get_behaviour()
            event = func(event, key)
        return event

    def unduplicate_durs(self, event):
        """ Converts values stored in event["dur"] in a tuple/PGroup into delays """

        # If there are more than one dur then add to the delay

        if "dur" in event:

            try:

                if len(event['dur']) > 1:

                    init_dur = event["dur"][0]

                    offset = PGroup(0|event['dur'][1:])

                    event["delay"] = event["delay"] + offset

                    event["dur"]   = float(init_dur)

                elif len(event['dur']) == 1:

                    event["dur"] = float(event["dur"][0])

            except TypeError:

                pass

        if "sus" in event:

            try:

                # Also update blur / sus

                if len(event['sus']) > 1:

                    min_sus = min(event['sus']) if min(event['sus']) else 1

                    offset = PGroup([(sus / min_sus) for sus in event["sus"]])

                    event["blur"] = event["blur"] * offset

                    event["sus"] = float(min_sus)

                elif len(event['sus']) == 1:

                    event["sus"] = float(event["sus"][0])

            except TypeError:

                pass

        return event

    def get_event(self):
        """ Returns a dictionary of attr -> now values """

        self.event = dict(map(lambda attr: (attr, self.now(attr)), self.attr.keys()))

        self.event = self.unduplicate_durs(self.event)

        self.event = self.get_prime_funcs(self.event)

        # Update internal player keys / schedule future updates

        self.update_all_player_keys()

        return self


    def send(self, timestamp=None, verbose=True, **kwargs):
        """ Goes through the  current event and compiles osc messages and sends to server via the tempo clock """

        timestamp = timestamp if timestamp is not None else self.queue_block.time

        # self.do_bang = False

        for i in range(self.get_event_length(**kwargs)):

            self.send_osc_message(self.event, i, timestamp=timestamp, verbose=verbose, **kwargs)

        # if self.do_bang:

        #     self.bang()

        return

    def send_osc_message(self, event, index, timestamp=None, verbose=True, **kwargs):
        """ Compiles and sends an individual OSC message created by recursively unpacking nested PGroups """

        packet = {}

        event=event.copy()
        event.update(kwargs)

        for key, value in event.items():

            # If we can index a value, trigger a new OSC message to send OSC messages for each

            # value = kwargs.get(key, value)

            if isinstance(value, PGroup):

                new_event = {}

                for new_key, new_value in event.items():

                    # new_value = kwargs.get(new_key, new_value)

                    if isinstance(new_value, PGroup):

                        new_event[new_key]  = new_value[index]

                    else:

                        new_event[new_key] = new_value

                # Recursively unpack and send messages

                for i in range(self.get_event_length(new_event)):

                    self.send_osc_message(new_event, i, timestamp, verbose)

                return

            else:

                # If it is a number, use the numbers (check for kwargs override)
                
                packet[key] = value

        # Special case modulations

        if ("amp" in packet) and ("amplify" in packet):

            packet["amp"] = packet["amp"] * packet["amplify"]

        # Send compiled messages

        self.push_osc_to_server(packet, timestamp, verbose, **kwargs)

        return

    def push_osc_to_server(self, packet, timestamp, verbose=True, **kwargs):
        """ Adds message head, calculating frequency then sends to server if verbose is True and 
            amp/bufnum values meet criteria """

        # Do any calculations e.g. frequency

        message = self.new_message_header(packet, **kwargs)

        # Only send if amp > 0 etc

        if verbose and (message["amp"] > 0) and ((self.synthdef != SamplePlayer and message["freq"] != None) or (self.synthdef == SamplePlayer and message["buf"] > 0)):

            # Need to send delay and synthdef separately

            delay = self.metro.beat_dur(message.get("delay", 0))

            synthdef = self.get_synth_name(message.get("buf", 0)) # to send to play1 or play2

            compiled_msg = self.metro.server.get_bundle(synthdef, message, timestamp = timestamp + delay)

            # We can set a condition to only send messages

            self.queue_block.append_osc_message(compiled_msg)

            # self.do_bang = True

        return

    def new_message_header(self, event, **kwargs):
        """ Returns the header of an osc message to be added to by osc_message() """

        # Let SC know the duration of 1 beat so effects can use it and adjust sustain too

        beat_dur = self.metro.beat_dur()

        message = {"beat_dur": beat_dur, "sus": kwargs.get("sus", event["sus"]) * beat_dur}

        if self.synthdef == SamplePlayer:

            degree = kwargs.get("degree", event['degree'])
            sample = kwargs.get("sample", event["sample"])
            rate   = kwargs.get("rate", event["rate"])

            if rate < 0:

                sus = kwargs.get("sus", event["sus"])

                pos = self.metro.beat_dur(sus)

            else:

                pos = 0 
 
            buf  = self.samples.getBufferFromSymbol(str(degree), sample).bufnum
            
            message.update( {'buf': buf,'pos': pos} )

            # Update player key

            if "buf" in self.accessed_keys:

                self.buf = buf

        elif self.synthdef == LoopPlayer:

            pos = kwargs.get("degree", event["degree"])
            buf = kwargs.get("buf", event["buf"])

            # Get a user-specified tempo

            given_tempo = kwargs.get("tempo", self.event.get("tempo", self.metro.bpm))

            if given_tempo in (None, 0):

                tempo = 1

            else:

                tempo = self.metro.bpm / given_tempo

            # Set the position in "beats"

            pos = pos * tempo * self.metro.beat_dur(1)

            # If there is a negative rate, move the pos forward

            rate = kwargs.get("rate", event["rate"])

            if rate == 0:

                rate = 1

            # Adjust the rate to a given tempo

            rate = float(tempo * rate)

            if rate < 0:

                sus = kwargs.get("sus", event["sus"])

                pos += self.metro.beat_dur(sus)

            message.update( {'pos': pos, 'buf': buf, 'rate': rate} )

        else:

            degree = kwargs.get("degree", event["degree"])
            octave = kwargs.get("oct", event["oct"])
            root   = kwargs.get("root", event["root"])

            scale  = kwargs.get("scale", self.scale)

            if degree == None:

                freq, midinote = None, None

            else:

                freq, midinote = get_freq_and_midi(degree, octave, root, scale)
                
            message.update({'freq':  freq, 'midinote': midinote})

            # Updater player key

            if "freq" in self.accessed_keys:

                self.freq = freq

            if "midinote" in self.accessed_keys:
    
               self.midinote = midinote

        # Update the dict with other values from the event

        event.update(message)

        # Remove keys we dont need

        del event["bpm"]
            
        return event        

    def set_queue_block(self, queue_block):
        """ Gives this player object a reference to the other items that are 
            scheduled at the same time """
        self.queue_block = queue_block
        return

    def get_synth_name(self, buf=0):
        """ Returns the real SynthDef name of the player. Useful only for "play" 
            as there is a play1 and play2 SynthDef for playing audio files with
            one or two channels respectively. """
        if self.synthdef == SamplePlayer:
            numChannels = self.samples.getBuffer(buf).channels
            if numChannels == 1:
                synthdef = "play1"
            else:
                synthdef = "play2"
        else:
            synthdef = str(self.synthdef)
        return synthdef

    def addfx(self, **kwargs):
        """ Not implemented - add an effect to the SynthDef bus on SuperCollider
            after it has been triggered. """
        return self

    #: Methods for stop/starting players

    def kill(self):
        """ Removes this object from the Clock and resets itself"""
        
        self.isplaying = False
        self.stopping = True
        
        self.reset()

        if self in self.metro.playing:
        
            self.metro.playing.remove(self)
        
        return
        
    def stop(self, N=0):
        
        """ Removes the player from the Tempo clock and changes its internal
            playing state to False in N bars time
            - When N is 0 it stops immediately"""

        self.stopping = True        
        self.stop_point = self.metro.now()

        if N > 0:

            self.stop_point = self.metro.next_bar() + ((N-1) * self.metro.bar_length())

        else:

            self.kill()

        return self

    def pause(self):

        self.isplaying = False

        return self

    def play(self):

        self.isplaying = True
        self.stopping = False
        self.isAlive = True

        self.__call__()

        return self

    # Methods for collaborative performance
    #
    # e.g. follow
    #

    def accompany(self, other, values=[0,2,4], debug=False):
        """ Similar to "follow" but when the value has changed """

        if isinstance(other, self.__class__):

            self.degree = other.degree.accompany()
        
        return self

    def follow(self, other=False):
        """ Takes a Player object and then follows the notes """

        if isinstance(other, self.__class__):

            self.degree = other.degree

        return self

    def versus(self, other_key, rule=lambda x, y: x > y, attr=None):
        """ Sets the 'amplify' key for both players to be dependent on the comparison of keys """

        # Get reference to the second player object

        other = other_key.player

        # Get the attribute from the key to versus

        this_key = getattr(self, other_key.attr if attr is None else attr)

        # Set amplifications based on the rule

        self.amplify  = this_key.transform(lambda value: rule(value, other_key.now()))
        other.amplify = this_key.transform(lambda value: not rule(value, other_key.now()))

        return self

    def reload(self):
        """ If this is a 'play' or 'loop' SynthDef, reload the filename used"""

        if self.synthdef == LoopPlayer:

            Samples.reload(self.filename)

        return self

    def only(self):
        """ Stops all players except this one """
        for player in list(self.metro.playing):
            if player is not self:
                player.stop()
        return self

    def solo(self, action=1):
        """ Silences all players except this player. Undo the solo
            by using `Player.solo(0)` """

        action=int(action)

        if action == 0:

            self.metro.solo.reset()

        elif action == 1:

            self.metro.solo.set(self)

        elif action == 2:

            pass

        return self

    def versus_old(self, other, key = lambda x: x.freq, f=max):
        """ Takes another Player object and a function that takes
            two player arguments and returns one, default is the higher
            pitched
        """
        if other is not None:
            assert(other.__class__ == self.__class__) # make sure it's using another player
            func = lambda x, y: f(x, y, key=key)
            self.condition  = lambda: func(self, other) == self
            other.condition = lambda: func(self, other) == other
            self._versus = other
        else:
            self.condition  = lambda: True
            self._versus.condition = lambda: True
            self._versus = None
        return self

    # def versus(self, other, func = lambda a, b: a > b):

    #     self.amp  = self.pitch > other.pitch
    #     other.amp = other.pitch > self.pitch

    #     return self
    

    # Utils

    def num_key_references(self):
        """ Returns the number of 'references' for the
            attr which references the most other players """
        num = 0
        for attr in self.attr.values():
            if isinstance(attr, PlayerKey):
                if attr.num_ref > num:
                    num = attr.num_ref
        return num

    def _replace_degree(self, new_degree):
        # Update the GUI if possible
        #if self.widget:
        #    if self.synthdef == SamplePlayer:
        #        if self.playstring is not None:
        #            # Replace old_string with new string (only works with plain string patterns)
        #            new_string = new_degree.string()
        #            self.widget.addTask(target=self.widget.replace, args=(self.line_number, self.playstring, new_string))
        #            self.playstring = new_string
        #    else:
        #        # Replaces the degree pattern in the widget (experimental)
        #        # self.widget.addTask(target=self.widget.replace_re, args=(self.line_number,), kwargs={'new':str(new_degree)})
        #        self.playstring = str(new_degree)
        setattr(self, 'degree', new_degree)
        return

    def multiply(self, n=2):
        self.attr['degree'] = self.attr['degree'] * n
        return self

    def degrade(self, amount=0.5):
        """ Sets the amp modifier to a random array of 0s and 1s
            amount=0.5 weights the array to equal numbers """
        if float(amount) <= 0:
            self.amplify = 1
        else:
            self.amplify = PwRand([0, self.attr["amplify"]],[int(amount*10), max(10 - int(amount),0)])
        return self

    def changeSynth(self, list_of_synthdefs):
        new_synth = choice(list_of_synthdefs)
        if isinstance(new_synth, SynthDef):
            new_synth = str(new_synth.name)
        self.synthdef = new_synth
        return self

    """

        Modifier Methods
        ----------------

        Other modifiers for affecting the playback of Players

    """

    def offbeat(self, dur=1):
        """ Off sets the next event occurence """

        self.dur = abs(dur)
        self.delay = abs(dur) / 2

        return self

    def strum(self, dur=0.025):
        """ Adds a delay to a Synth Envelope """
        x = self.largest_attribute()
        if x > 1:
            self.delay = asStream([tuple(a * dur for a in range(x))])
        else:
            self.delay = asStream(dur)
        return self

    def __repr__(self):
        if self.id is not None:
            return "<{} - {}>".format(self.id, self.synthdef)
        else:
            return "a '{}' Player Object".format(self.synthdef)

    def info(self):
        s = "Player Instance using '%s' \n\n" % self.synthdef
        s += "ATTRIBUTES\n"
        s += "----------\n\n"
        for attr, val in self.attr.items():
            s += "\t{}\t:{}\n".format(attr, val)
        return s

    def bang(self, **kwargs):
        """
        Triggered when sendNote is called. Responsible for any
        action to be triggered by a note being played. Default action
        is underline the player
        """
        if kwargs:

            self.bang_kwargs = kwargs

        elif self.bang_kwargs:

            bang = Bang(self, self.bang_kwargs)

        return self
        

###### GROUP OBJECT

class Group:

    metro = None

    def __init__(self, *args):

        self.players = list(args)

    def add(self, other):
        self.players.append(other)

    def __len__(self):
        return len(self.players)

    def __str__(self):
        return str(self.players)

    def solo(self, arg=True):

        if self.metro is None:

            self.__class__.metro = Player.metro

        if arg:
            
            self.metro.solo.set(self.players[0])

            for player in self.players[1:]:

                self.metro.solo.add(player)

        else:

            self.metro.solo.reset()

        return self
        
    def only(self):

        if self.metro is None:

            self.__class__.metro = Player.metro
        
        for player in list(self.metro.playing):
            
            if player not in self.players:
                
                player.stop()

        return self

    def iterate(self, dur=4):
        if dur == 0 or dur is None:
            self.amplify=1
        else:
            delay, on = 0, float(dur) / len(self.players)
            for player in self.players:
                player.amplify=TimeVar([0,1,0],[delay, on, dur-delay])
                delay += on
        return           

    def __setattr__(self, name, value):
        try:
            for p in self.players:
                try:
                    setattr(p, name, value)
                except:
                    WarningMsg("'%s' object has no attribute '%s'" % (str(p), name))
        except KeyError :
            self.__dict__[name] = value 
        return self        

    def __getattr__(self, name):
        """ Returns a Pattern object containing the desired attribute for each player in the group  """
        if name == "players":
            return self.__dict__["players"]
        attributes = GroupAttr()
        for player in self.players:
            if hasattr(player, name):
                attributes.append(getattr(player, name))
        return attributes

class GroupAttr(list):
    def __call__(self, *args, **kwargs):
        for p in self:
            if callable(p):
                p.__call__(*args, **kwargs)

class rest(object):
    ''' Represents a rest when used with a Player's `dur` keyword
    '''
    def __init__(self, dur=1):
        self.dur = dur if not isinstance(dur, self.__class__) else dur.dur
    def __repr__(self):
        return "<rest: {}>".format(self.dur)
    def __add__(self, other):
        return rest(self.dur + other)
    def __radd__(self, other):
        return rest(other + self.dur)
    def __sub__(self, other):
        return rest(self.dur - other)
    def __rsub__(self, other):
        return rest(other - self.dur)
    def __mul__(self, other):
        return rest(self.dur * other)
    def __rmul__(self, other):
        return rest(other * self.dur)
    def __div__(self, other):
        return rest(self.dur / other)
    def __rdiv__(self, other):
        return rest(other / self.dur)
    def __truediv__(self, other):
        return rest(float(self.dur) / other)
    def __rtruediv__(self, other):
        return rest(other / float(self.dur))
    def __mod__(self, other):
        return rest(self.dur % other)
    def __rmod__(self, other):
        return rest(other % self.dur)
    def __eq__(self, other):
        return (self.dur == other)
    def __ne__(self, other):
        return (self.dur != other)
    def __lt__(self, other):
        return (self.dur < other)
    def __le__(self, other):
        return (self.dur <= other)
    def __gt__(self, other):
        return (self.dur > other)
    def __ge__(self, other):
        return (self.dur >= other)
    def __int__(self):
        return int(self.dur)
    def __float__(self):
        return float(self.dur)

class PlayerKeyException(Exception):
    pass
